"""Module shell

Defines the shell to be used in pyzo.
This is done in a few inheritance steps:
  - BaseShell inherits BaseTextCtrl and adds the typical shell behaviour.
  - PythonShell makes it specific to Python.
This module also implements ways to communicate with the shell and to run
code in it.

"""

import sys
import time
import re

import yoton

import pyzo
from pyzo.util import zon as ssdf  # zon is ssdf-light
from pyzo.qt import QtCore, QtGui, QtWidgets

Qt = QtCore.Qt

from pyzo.codeeditor.highlighter import Highlighter
from pyzo.codeeditor import parsers

from pyzo.core.baseTextCtrl import BaseTextCtrl
from pyzo.core.pyzoLogging import print
from pyzo.core.kernelbroker import KernelInfo, Kernelmanager
from pyzo.core.menu import ShellContextMenu

ismacos = sys.platform.startswith("darwin")

# Interval for polling messages. Timer for each kernel.
POLL_TIMER_INTERVAL_IDLE = 100  # in ms; energy saving mode
POLL_TIMER_INTERVAL_BUSY = 10  # in ms; faster updates

# Maximum number of lines in the shell
MAXBLOCKCOUNT = pyzo.config.advanced.shellMaxLines


# todo: we could make command shells to, with autocompletion and coloring...


class YotonEmbedder(QtCore.QObject):
    """Embed the Yoton event loop."""

    def __init__(self):
        super().__init__()
        yoton.app.embed_event_loop(self.postYotonEvent)

    def postYotonEvent(self):
        try:
            QtWidgets.qApp.postEvent(self, QtCore.QEvent(QtCore.QEvent.Type.User))
        except Exception:
            pass  # If pyzo is shutting down, the app may be None

    def customEvent(self, event):
        """This is what gets called by Qt."""
        yoton.process_events(False)


yotonEmbedder = YotonEmbedder()


# Short constants for cursor movement
A_KEEP = QtGui.QTextCursor.MoveMode.KeepAnchor
A_MOVE = QtGui.QTextCursor.MoveMode.MoveAnchor

# Instantiate a local kernel broker upon loading this module
pyzo.localKernelManager = Kernelmanager(public=False)


def finishKernelInfo(info, scriptFile=None):
    """Get a copy of the kernel info struct, with the scriptFile
    and the projectPath set.
    """

    # Make a copy, we do not want to change the original
    info = ssdf.copy(info)

    # Set scriptFile (if '', the kernel will run in interactive mode)
    if scriptFile:
        info.scriptFile = scriptFile
    else:
        info.scriptFile = ""

    # If the file browser is active, and has the check box
    #'add path to Python path' set, set the PROJECTPATH variable
    fileBrowser = pyzo.toolManager.getTool("pyzofilebrowser")
    info.projectPath = ""
    if fileBrowser:
        info.projectPath = fileBrowser.getAddToPythonPath()

    return info


class ShellHighlighter(Highlighter):
    """This highlighter implements highlighting for a shell;
    only the input lines are highlighted with this highlighter.
    """

    def highlightBlock(self, line):
        # Get previous state
        previousState = self.previousBlockState()

        # Get parser
        parser = None
        if hasattr(self._codeEditor, "parser"):
            parser = self._codeEditor.parser()

        # Get function to get format
        nameToFormat = self._codeEditor.getStyleElementFormat

        # Last line?
        cursor1 = self._codeEditor._cursor1
        cursor2 = self._codeEditor._cursor2
        commandCursor = self._codeEditor._lastCommandCursor
        curBlock = self.currentBlock()
        #
        atLastPrompt, atCurrentPrompt = False, False
        if curBlock.position() == 0:
            pass
        elif curBlock.position() == commandCursor.block().position():
            atLastPrompt = True
        elif curBlock.position() >= cursor1.block().position():
            atCurrentPrompt = True

        if not atLastPrompt and not atCurrentPrompt:
            # Do not highlight anything but current and last prompts
            return

        # Get user data
        bd = self.getCurrentBlockUserData()

        if parser:
            if atCurrentPrompt:
                pos1, pos2 = cursor1.positionInBlock(), cursor2.positionInBlock()
            else:
                pos1, pos2 = 0, commandCursor.positionInBlock()

            # Check if we should *not* format this line.
            # This is the case for special "executing" text
            # A bit of a hack though ... is there a better way to signal this?
            specialinput = (not atCurrentPrompt) and line[pos2:].startswith(
                "(executing "
            )

            self.setCurrentBlockState(0)
            if specialinput:
                pass  # Let the kernel decide formatting
            else:
                tokens = parser.parseLine(line, previousState)
                bd.tokens = tokens
                for token in tokens:
                    # Handle block state
                    if isinstance(token, parsers.BlockState):
                        self.setCurrentBlockState(token.state)
                    else:
                        # Get format
                        try:
                            format = nameToFormat(token.name).textCharFormat
                        except KeyError:
                            # print(repr(nameToFormat(token.name)))
                            continue
                        # Set format
                        # format.setFontWeight(QtGui.QFont.Weight.Bold)
                        if token.start >= pos2:
                            self.setFormat(token.start, token.end - token.start, format)

            # Set prompt to bold
            if atCurrentPrompt:
                format = QtGui.QTextCharFormat()
                format.setFontWeight(QtGui.QFont.Weight.Bold)
                self.setFormat(pos1, pos2 - pos1, format)

        # Get the indentation setting of the editors
        indentUsingSpaces = self._codeEditor.indentUsingSpaces()

        leadingWhitespace = line[: len(line) - len(line.lstrip())]
        if "\t" in leadingWhitespace and " " in leadingWhitespace:
            # Mixed whitespace
            bd.indentation = 0
            format = QtGui.QTextCharFormat()
            format.setUnderlineStyle(
                QtGui.QTextCharFormat.UnderlineStyle.SpellCheckUnderline
            )
            format.setUnderlineColor(QtCore.Qt.GlobalColor.red)
            format.setToolTip("Mixed tabs and spaces")
            self.setFormat(0, len(leadingWhitespace), format)
        elif ("\t" in leadingWhitespace and indentUsingSpaces) or (
            " " in leadingWhitespace and not indentUsingSpaces
        ):
            # Whitespace differs from document setting
            bd.indentation = 0
            format = QtGui.QTextCharFormat()
            format.setUnderlineStyle(
                QtGui.QTextCharFormat.UnderlineStyle.SpellCheckUnderline
            )
            format.setUnderlineColor(QtCore.Qt.GlobalColor.blue)
            format.setToolTip("Whitespace differs from document setting")
            self.setFormat(0, len(leadingWhitespace), format)
        else:
            # Store info for indentation guides
            # amount of tabs or spaces
            bd.indentation = len(leadingWhitespace)


class BaseShell(BaseTextCtrl):
    """The BaseShell implements functionality to make a generic shell."""

    SHORTENED_LINE_MARKER = "\u200b"  # '\u200b' ... zero width space

    def __init__(self, parent, **kwds):
        super().__init__(
            parent,
            wrap=True,
            showLineNumbers=False,
            showBreakPoints=False,
            highlightCurrentLine=False,
            parser="python",
            **kwds,
        )

        # Use a special highlighter that only highlights the input.
        self._setHighlighter(ShellHighlighter)

        # No undo in shells
        self.setUndoRedoEnabled(False)

        # variables we need
        self._more = False

        # We use two cursors to keep track of where the prompt is
        # cursor1 is in front, and cursor2 is at the end of the prompt.
        # They can be in the same position.
        # Further, we store a cursor that selects the last given command,
        # so it can be styled.
        self._cursor1 = self.textCursor()
        self._cursor2 = self.textCursor()
        self._lastCommandCursor = self.textCursor()
        self._lastline_had_cr = False
        self._lastline_had_lf = False

        # When inserting/removing text at the edit line (thus also while typing)
        # keep cursor2 at its place. Only when text is written before
        # the prompt (i.e. in write), this flag is temporarily set to False.
        # Same for cursor1, because sometimes (when there is no prompt) it
        # is at the same position.
        self._cursor1.setKeepPositionOnInsert(True)
        self._cursor2.setKeepPositionOnInsert(True)

        # Similarly, we use the _lastCommandCursor cursor really for pointing.
        self._lastCommandCursor.setKeepPositionOnInsert(True)

        self.resetShellWriters()

        # Variables to keep track of the command history usage
        self._historyNeedle = None  # None means none, "" means look in all
        self._historyStep = 0

        # Set minimum width so 80 lines do fit in smallest font size
        self.setMinimumWidth(200)

        # Hard wrapping. QTextEdit allows hard wrapping at a specific column.
        # Unfortunately, QPlainTextEdit does not.
        self.setWordWrapMode(QtGui.QTextOption.WrapMode.WrapAnywhere)

        # Limit number of lines
        self.setMaximumBlockCount(MAXBLOCKCOUNT)

        # Keep track of position, so we can disable editing if the cursor
        # is before the prompt
        self.cursorPositionChanged.connect(self.onCursorPositionChanged)

        self.setFocusPolicy(Qt.FocusPolicy.TabFocus)  # See remark at mousePressEvent

    ## Cursor stuff

    def onCursorPositionChanged(self):
        # If the end of the selection (or just the cursor if there is no selection)
        # is before the beginning of the line. make the document read-only
        cursor = self.textCursor()
        promptpos = self._cursor2.position()
        if cursor.position() < promptpos or cursor.anchor() < promptpos:
            self.setReadOnly(True)
        else:
            self.setReadOnly(False)

    def ensureCursorAtEditLine(self):
        """If the text cursor is before the beginning of the edit line,
        move it to the end of the edit line
        """
        cursor = self.textCursor()
        promptpos = self._cursor2.position()
        if cursor.position() < promptpos or cursor.anchor() < promptpos:
            cursor.movePosition(
                cursor.MoveOperation.End, A_MOVE
            )  # Move to end of document
            self.setTextCursor(cursor)
            self.onCursorPositionChanged()

    def mousePressEvent(self, event):
        """
        - Focus policy
            If a user clicks this shell, while it has no focus, we do
            not want the cursor position to change (since generally the
            user clicks the shell to give it the focus). We do this by
            setting the focus-policy to Qt::TabFocus, and we give the
            widget its focus manually from the mousePressedEvent event
            handler
        """
        if not self.hasFocus():
            self.setFocus()
            return
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.MouseButton.MiddleButton:
            if sys.platform == "linux":
                # On Linux, pasting with the middle button would insert text at the
                # clicked position, even if this is before the prompt (if readOnly is False
                # when the current text cursor is at the edit line).
                # We will ignore the pasting if the click position is not after the prompt.
                cursor = self.cursorForPosition(event.position().toPoint())
                promptpos = self._cursor2.position()
                if cursor.position() < promptpos:
                    return
                else:
                    # The current text cursor might be before the end of the prompt.
                    # But the text is inserted after the prompt (mouse event position).
                    # Therefore we move the text cursor to the edit line so that we are
                    # not in read-only mode.
                    # The text will be inserted at the event's position, no matter where
                    # the text cursor is.
                    self.setTextCursor(cursor)
                    self.onCursorPositionChanged()
        super().mouseReleaseEvent(event)

    def contextMenuEvent(self, event):
        """Do not show context menu."""
        pass

    def mouseDoubleClickEvent(self, event):
        super().mouseDoubleClickEvent(event)
        self._handleClickOnFilename(event.position().toPoint())

    def _handleClickOnFilename(self, mousepos):
        """Check whether the text that is clicked is a filename
        and open the file in the editor. If a line number can also be
        detected, open the file at that line number.
        """

        # Get cursor and its current pos
        cursor = self.cursorForPosition(mousepos)
        ocursor = QtGui.QTextCursor(cursor)  # Make a copy to use below
        pos = cursor.positionInBlock()

        # Define some abbreviations for movePosition arguments
        MO = cursor.MoveOperation
        MM = cursor.MoveMode

        # Get the whole line where the double click occurred
        cursor.movePosition(MO.EndOfBlock, MM.MoveAnchor)
        cursor.movePosition(MO.StartOfBlock, MM.KeepAnchor)
        line = cursor.selectedText()
        if len(line) > 1024:
            return  # safety

        if sys.platform == "win32":
            # e.g. "C:\somefile" or "\\abc" or "c:/abc"
            patFile = r"(?:\\\\|[a-zA-Z]:[\\/]).+?"
        else:
            patFile = r"/.+?"

        patFileOrTmp = r"(" + patFile + r"|<tmp \d+>)"

        # There could be a line offset after the filename, e.g.:
        # <tmp 3>+2:3: SyntaxWarning: invalid escape sequence '\s'
        patFileOrTmpWithOffset = patFileOrTmp + r"(\+\d+)?"  # groups: filepath, offset

        filename = None
        offset = None
        linenr = None

        patWarning = patFileOrTmpWithOffset + r":(\d+): [a-zA-Z]+Warning: .*"
        # /tmp/aa.py+5:4: SyntaxWarning: invalid escape sequence '\s'

        patIPython = r"\s*File " + patFileOrTmpWithOffset + r":(\d+)(?:, in .*)?"
        # IPython examples (leading spaces occur when running file as script):
        # File /tmp/aa.py:12
        # File /tmp/aa.py:4, in myfunc1()
        #   File /tmp/aa.py:14

        for pattern in [patWarning, patIPython]:
            mo = re.fullmatch(pattern, line)
            if mo:
                i1, i2 = mo.span(1)
                if i1 <= pos < i2:
                    filename = mo[1]
                    offset = mo[2]
                    linenr = int(mo[3])
                    break
                else:
                    # the pattern matches, but the double click was outside the filepath
                    return

        if filename is None:
            # Expand string to left and right, starting at the clicked position, and
            # stopping only when reaching the ends or encountering a (single) quote.
            line2 = line.replace("'", '"')
            i1 = line2.rfind('"', 0, pos) + 1
            i2 = line2.find('"', pos)
            if i2 == -1:
                i2 = len(line)

            mo = re.fullmatch(patFileOrTmpWithOffset, line[i1:i2])
            if mo:
                filename, offset = mo.groups()

                # Split in parts for getting line number
                mo = re.search(r"\b(?:line|linenr|lineno)\b\s+(\d+)", line[i2:])
                if mo:
                    linenr = int(mo[1])

        if filename:
            if offset and offset.startswith("+") and linenr is not None:
                linenr += int(offset[1:])

            # Select the whole filename in the shell
            cursor = ocursor
            cursor.movePosition(MO.Left, MM.MoveAnchor, pos - i1)
            cursor.movePosition(MO.Right, MM.KeepAnchor, len(filename))
            self.setTextCursor(cursor)

            # Try opening the file (at the line number if we have one)
            result = pyzo.editors.loadFile(filename)
            if result:
                editor = result._editor
                if linenr is not None:
                    editor.gotoLine(linenr)
                    cursor = editor.textCursor()
                    cursor.movePosition(MO.StartOfBlock)
                    cursor.movePosition(MO.EndOfBlock, MM.KeepAnchor)
                    editor.setTextCursor(cursor)
                editor.setFocus()

    ## Indentation: override code editor behaviour
    def indentSelection(self):
        pass

    def dedentSelection(self):
        pass

    ## Key handlers
    def keyPressEvent(self, event):
        if event.key() in [Qt.Key.Key_Return, Qt.Key.Key_Enter]:
            # First check if autocompletion triggered
            if self.potentiallyAutoComplete(event):
                return
            else:
                # Enter: execute line
                # Remove calltip and autocomp if shown
                self.autocompleteCancel()
                self.calltipCancel()

                # reset history needle
                self._historyNeedle = None

                # process
                self.processLine()
                return

        if event.key() == Qt.Key.Key_Escape:
            # Escape clears command
            if not (self.autocompleteActive() or self.calltipActive()):
                self.clearCommand()

        if (
            event.key() in (Qt.Key.Key_PageUp, Qt.Key.Key_PageDown)
            and not self.autocompleteActive()
        ):
            # The cursor should not move. So we have to ignore the event.
            # But we still want to be able to scroll the shell.
            # --> We send the event directly to the base class so
            #     that the extensions do not see the event.
            pyzo.codeeditor.CodeEditorBase.keyPressEvent(self, event)
            return

        if event.key() == Qt.Key.Key_Home:
            # Home goes to the prompt.
            cursor = self.textCursor()
            if event.modifiers() & Qt.KeyboardModifier.ShiftModifier:
                cursor.setPosition(self._cursor2.position(), A_KEEP)
            else:
                cursor.setPosition(self._cursor2.position(), A_MOVE)
            #
            self.setTextCursor(cursor)
            self.autocompleteCancel()
            return

        if event.key() == Qt.Key.Key_Insert:
            # Don't toggle between insert mode and overwrite mode.
            return True

        # Ensure to not backspace / go left beyond the prompt
        if event.key() == Qt.Key.Key_Left:
            self._historyNeedle = None
            if self.textCursor().position() == self._cursor2.position():
                return  # Ignore the key, don't go beyond the prompt

        # handle backspace
        if event.key() == Qt.Key.Key_Backspace:
            cursor = self.textCursor()
            self._historyNeedle = None
            if cursor.hasSelection():
                if cursor.selectionStart() < self._cursor2.position():
                    # The selection contains read-only text before or including the prompt.
                    # We do not want to delete that.
                    return  # ignore the key
            elif cursor.position() <= self._cursor2.position():
                return  # Ignore the key, don't go beyond the prompt
            else:
                KM = Qt.KeyboardModifier
                controlPressed = event.modifiers() == (
                    KM.AltModifier if ismacos else KM.ControlModifier
                )
                if controlPressed:
                    self._historyNeedle = None
                    cursor = self.textCursor()
                    text = cursor.block().text()[
                        self._cursor2.positionInBlock() : cursor.positionInBlock()
                    ]
                    # Ctrl+Backspace should not delete our prompt. This would happen if
                    # someone presses Ctrl+Backspace after only entering whitespace at
                    # the prompt.
                    if False:
                        # this works for any prompt, even without any whitespace at the end:
                        mo = re.match(r"(\s*([^\s\W]+|[^\s\w]+|))", text[::-1])
                        numDelete = len(mo[1])
                        cursor.movePosition(
                            cursor.MoveOperation.Left, A_KEEP, numDelete
                        )
                        cursor.removeSelectedText()
                        return
                    else:
                        # a bit simpler, but requires a space at the end of the prompt:
                        if text.strip() == "":
                            numDelete = len(text)
                            cursor.movePosition(
                                cursor.MoveOperation.Left, A_KEEP, numDelete
                            )
                            cursor.removeSelectedText()
                            return

        if (
            event.key() in [Qt.Key.Key_Up, Qt.Key.Key_Down]
            and not self.autocompleteActive()
        ):
            # needle
            if self._historyNeedle is None:
                # get partly-written-command
                #
                # Select text
                cursor = self.textCursor()
                cursor.setPosition(self._cursor2.position(), A_MOVE)
                cursor.movePosition(cursor.MoveOperation.End, A_KEEP)
                # Update needle text
                self._historyNeedle = cursor.selectedText()
                self._historyStep = 0

            # Browse through history
            if event.key() == Qt.Key.Key_Up:
                self._historyStep += 1
            else:  # Key_Down
                self._historyStep -= 1
                if self._historyStep < 1:
                    self._historyStep = 1

            # find the command
            c = pyzo.command_history.find_starting_with(
                self._historyNeedle, self._historyStep
            )
            if c is None:
                # found nothing-> reset
                self._historyStep = 0
                c = self._historyNeedle

            # Replace text
            cursor = self.textCursor()
            cursor.setPosition(self._cursor2.position(), A_MOVE)
            cursor.movePosition(cursor.MoveOperation.End, A_KEEP)
            cursor.insertText(c)

            self.ensureCursorAtEditLine()
            return

        else:
            # Reset needle
            self._historyNeedle = None

        # if a 'normal' key is pressed, ensure the cursor is at the edit line
        if event.text():
            self.ensureCursorAtEditLine()

        # Default behaviour: BaseTextCtrl
        super().keyPressEvent(event)

    ## Cut / Copy / Paste / Drag & Drop

    def cut(self):
        """Reimplement cut to only copy if part of the selected text
        is not at the prompt.
        """

        if self.isReadOnly():
            return self.copy()
        else:
            return super().cut()

    # def copy(self): # no overload needed

    def paste(self):
        """Reimplement paste to paste at the end of the edit line when
        the position is at the prompt.
        """
        self.ensureCursorAtEditLine()
        # Paste normally
        return super().paste()

    def dragEnterEvent(self, event):
        """We only support copying of the text"""
        if event.mimeData().hasText():
            event.setDropAction(QtCore.Qt.DropAction.CopyAction)
            event.accept()

    def dragMoveEvent(self, event):
        self.dragEnterEvent(event)

    def dropEvent(self, event):
        """The shell supports only a single line but the text may contain multiple
        lines. We insert at the editLine only the first non-empty line of the text
        """
        if event.mimeData().hasText():
            text = event.mimeData().text()
            insertText = ""
            for line in text.splitlines():
                if line.strip():
                    insertText = line
                    break

            # Move the cursor to the position indicated by the drop location, but
            # ensure it is at the edit line
            self.setTextCursor(self.cursorForPosition(event.position().toPoint()))
            self.ensureCursorAtEditLine()

            # Now insert the text
            cursor = self.textCursor()
            cursor.insertText(insertText)
            self.setFocus()

    def createMimeDataFromSelection(self):
        SLM = self.SHORTENED_LINE_MARKER
        mime = QtCore.QMimeData()
        # We keep this code simple because it is executed every time the text selection
        # in the widget is changed.
        t = (
            self.textCursor()
            .selectedText()
            .replace("\u2029", "\n")
            .replace(SLM + "\n", "")
        )
        if t.endswith(SLM):
            t = t[: -len(SLM)]
        mime.setText(t)
        return mime

    ## Basic commands to control the shell

    def clearScreen(self):
        """Clear all the previous output from the screen."""
        # Select from beginning of prompt to start of document
        self._cursor1.clearSelection()
        self._cursor1.movePosition(
            self._cursor1.MoveOperation.Start, A_KEEP
        )  # Keep anchor
        self._cursor1.removeSelectedText()
        # Wrap up
        self.ensureCursorAtEditLine()
        self.ensureCursorVisible()
        self._cursor1_lastStreamStart = None

    def deleteLines(self):
        """Called from the menu option "delete lines", just execute self.clearCommand()"""
        self.clearCommand()

    def clearCommand(self):
        """Clear the current command, move the cursor right behind
        the prompt, and ensure it's visible.
        """
        # Select from prompt end to length and delete selected text.
        cursor = self.textCursor()
        cursor.setPosition(self._cursor2.position(), A_MOVE)
        cursor.movePosition(cursor.MoveOperation.End, A_KEEP)
        cursor.removeSelectedText()
        # Wrap up
        self.ensureCursorAtEditLine()
        self.ensureCursorVisible()

    def write(self, text, prompt=0, color=None, streamIdentifier=None):
        """Write to the shell.

        If prompt is 0 (default) the text is printed before the prompt. If
        prompt is 1, the text is printed after the prompt, the new prompt
        becomes null. If prompt is 2, the given text becomes the new prompt.

        The color of the text can also be specified (as a hex-string).

        The streamIdentifier is a hash-able value that serves to recognize the
        stream, especially when we are writing before the prompt. For example:
        When we print 'SOMETEXT' to stream stdout it might be split into
        smaller fragments, e.g. 'SOME' and 'TEXT'. Between writing these two
        fragments to the shell widget, some other text 'error' might be printed
        to stream stderr. So the final output could be 'SOMEerrorTEXT'.
        When writing these three strings we have to know if they are from the
        same stream or not. This is important when a format escape segment is
        split into two fragments or when a backspace character wants to delete
        a character from the previous fragment of the same stream.
        """

        # From The Qt docs: Note that a cursor always moves when text is
        # inserted before the current position of the cursor, and it always
        # keeps its position when text is inserted after the current position
        # of the cursor.

        # Make sure there's text and make sure it's a string
        if not text:
            return
        if isinstance(text, bytes):
            text = text.decode("utf-8")

        # Prepare format
        format = QtGui.QTextCharFormat()
        if color:
            format.setForeground(QtGui.QColor(color))

        # pos1, pos2 = self._cursor1.position(), self._cursor2.position()

        # Just in case, clear any selection of the cursors
        self._cursor1.clearSelection()
        self._cursor2.clearSelection()

        if prompt == 0:
            # Insert text before prompt (normal streams)
            self._cursor1.setKeepPositionOnInsert(False)
            self._cursor2.setKeepPositionOnInsert(False)

            # We want to know if we are allowed to remove some characters from the
            # current line (left to the cursor), and if yes, till which column.
            # If the current line started with a different stream, we cannot delete
            # from that.
            leftLimit = None
            if self._cursor1_lastStreamStart is not None:
                if self._cursor1_lastStreamStart[0] == streamIdentifier:
                    leftLimit = self._cursor1_lastStreamStart[1]

            shellWriter = self._shellWriters.setdefault(streamIdentifier, ShellWriter())
            newLeftLimit = shellWriter.writeText(self._cursor1, leftLimit, text, format)
            self._cursor1_lastStreamStart = (streamIdentifier, newLeftLimit)
        elif prompt == 1:
            # Insert command text after prompt, prompt becomes null (input)
            self._lastCommandCursor.setPosition(self._cursor2.position())
            self._cursor1.setKeepPositionOnInsert(False)
            self._cursor2.setKeepPositionOnInsert(False)
            ShellWriter().writeText(self._cursor2, None, text, format)
            self._cursor1.setPosition(self._cursor2.position(), A_MOVE)
            self._cursor1_lastStreamStart = None
        elif prompt == 2 and text == "\b":
            # Remove prompt (used when closing the kernel)
            self._cursor1.setPosition(self._cursor2.position(), A_KEEP)
            self._cursor1.removeSelectedText()
            self._cursor2.setPosition(self._cursor1.position(), A_MOVE)
        elif prompt == 2:
            # text becomes new prompt
            self._cursor1.setPosition(self._cursor2.position(), A_KEEP)
            self._cursor1.removeSelectedText()
            self._cursor1.setKeepPositionOnInsert(True)
            self._cursor2.setKeepPositionOnInsert(False)
            ShellWriter().writeText(self._cursor1, None, text, format)

        # Reset cursor states for the user to type his/her commands
        self._cursor1.setKeepPositionOnInsert(True)
        self._cursor2.setKeepPositionOnInsert(True)

        # Make sure that cursor is visible (only when cursor is at edit line)
        if not self.isReadOnly():
            self.ensureCursorVisible()

        # Scroll along with the text if lines are popped from the top
        elif self.blockCount() == MAXBLOCKCOUNT:
            n = text.count("\n")
            sb = self.verticalScrollBar()
            sb.setValue(sb.value() - n)

    ## Executing stuff

    def processLine(self, line=None, execute=True):
        """Process the given line or the current line at the prompt if not given.

        Called when the user presses enter.

        If execute is False will not execute the command. This way
        a message can be written while other ways are used to process
        the command.
        """

        # Can we do this?
        if self.isReadOnly() and not line:
            return

        if line:
            # remove trailing newline(s)
            command = line.rstrip("\n")
        else:
            # Select command
            cursor = self.textCursor()
            cursor.setPosition(self._cursor2.position(), A_MOVE)
            cursor.movePosition(cursor.MoveOperation.End, A_KEEP)

            # Sample the text from the prompt and remove it
            command = cursor.selectedText().replace("\u2029", "\n").rstrip("\n")
            cursor.removeSelectedText()

            # Auto-indent. Note: this is rather Python-specific
            command_s = command.lstrip()
            indent = " " * (len(command) - len(command_s))
            if command.strip().endswith(":"):
                indent += "    "
            elif not command_s:
                indent = ""
            if indent:
                cursor.insertText(indent)

            if command:
                # Remember the command in this global history
                pyzo.command_history.append(command)

        if execute:
            command = command.replace("\r\n", "\n")
            self.executeCommand(command + "\n")

    def executeCommand(self, command):
        """Execute the given command.
        Should be overridden.
        """
        # this is a stupid simulation version
        self.write("you executed: " + command + "\n")
        self.write(">>> ", prompt=2)

    def resetShellWriters(self):
        # init shell writer objects (for handling terminal escape sequences etc.)
        self._shellWriters = {}
        self._cursor1_lastStreamStart = None


class ShellWriter:
    # normal colors:
    # COLORS = "#000 #F00 #0F0 #FF0 #00F #F0F #0FF #FFF".split()

    # solarised color theme:
    COLORS = "#657b83 #dc322f #859900 #b58900 #268bd2 #d33682 #2aa198 #eee8d5".split()

    _linebreaks = "\n\u2028\u2029\ufdd0\ufdd1"
    _reSplit = re.compile("([" + _linebreaks + "\t\x1b]|\v+|\r+|\b+)")
    _reFormatPattern = re.compile(r"(\x1b\[(\d+(?:;\d+)*)m)")

    def __init__(self):
        self._currentFormat = None
        self._unfinishedTail = ""  # either "" or "\r" or "\x1b"+...
        self._lineShorteningActive = False

    def writeText(self, cursor, leftLimitFirstRow, text, defaultFormat):
        if leftLimitFirstRow is None:
            leftLimitFirstRow = cursor.positionInBlock()
            self._lineShorteningActive = False

        assert cursor.positionInBlock() >= leftLimitFirstRow

        leftLimit = leftLimitFirstRow
        posInBlock = cursor.positionInBlock()
        lowestPosInFirstBlock = posInBlock
        unfinishedTailNew = ""

        text2 = text
        if self._unfinishedTail.startswith("\x1b"):
            posInBlock = max(0, posInBlock - len(self._unfinishedTail))
            lowestPosInFirstBlock = posInBlock
            text2 = self._unfinishedTail + text

        if self._currentFormat is None:
            self._currentFormat = defaultFormat
        currentFormat = self._currentFormat

        finishedLines = []
        # each element in finishedLines is a list like this for each line:
        #   [*elems, endPos]
        #   0 to n elems ... either of type str or QTextCharFormat
        #   endPos ... end column (0-based), after the last character
        #
        #   Format elems have length zero.
        #   The last elem (before endPos) can be a str that ends with a '\n'.
        #   There is no other '\n' anywhere else in the line.

        ll = []
        splitText = [s for s in self._reSplit.split(text2) if s != ""]
        if self._unfinishedTail == "\r" and splitText[:1] != ["\n"]:
            # the last text ended with '\r', and it is not a '\r\n'
            # --> clear this line
            splitText.insert(0, "\r")
        lenSplit = len(splitText)
        for i, s in enumerate(splitText):
            c = s[:1]
            if c == "":
                continue
            elif c in self._linebreaks:  # new line
                s = "\n"
                ll.append(s)
                posInBlock += len(s)
                ll.append(posInBlock)
                finishedLines.append(ll)
                ll = []
                leftLimit = 0
                posInBlock = leftLimit
            elif c == "\t":  # horizontal tab --> moves to next multiple of tabwidth
                tabwidth = 8
                numSpaces = tabwidth - posInBlock % tabwidth
                s = " " * numSpaces
                posInBlock += len(s)
                ll.append(s)
            elif c == "\v":  # vertical tab --> moves to next line, but same column
                n = len(s)
                ll.append("\n")
                ll.append(posInBlock + 1)
                finishedLines.append(ll)
                leftLimit = 0
                finishedLines.extend([["\n", leftLimit + 1]] * (n - 1))
                ll = [" " * posInBlock]
            elif c == "\r":
                if i == lenSplit - 1:
                    # we don't know yet if this is a CR LF or a CR to delete the line
                    unfinishedTailNew = c
                    continue
                elif splitText[i + 1] == "\n":
                    # we have a CR LF sequence --> ignore the CR
                    continue
                else:
                    # delete the current line
                    ll = []
                    posInBlock = leftLimit
                    if len(finishedLines) == 0:
                        lowestPosInFirstBlock = leftLimit
            elif c == "\b":  # backspace
                # Backspace normally would move the cursor one character left without
                # deleting anything. In our implementation, backspace removes the
                # previous character, even if it was from a text fragment of the same
                # stream that was already written before. We can only delete till the
                # start of the current line, and we cannot move further left than
                # column leftLimit.
                n = len(s)
                for i, v in list(enumerate(ll))[::-1]:
                    if isinstance(v, str):
                        numBackspace = min(n, len(v))
                        ll[i] = v[: len(v) - numBackspace]
                        n -= numBackspace
                        posInBlock -= numBackspace
                        if n == 0:
                            break
                posInBlock = max(leftLimit, posInBlock - n)
                if len(finishedLines) == 0 and posInBlock < lowestPosInFirstBlock:
                    lowestPosInFirstBlock = posInBlock
            elif c == "\x1b":  # could be the start of an ANSI-like escape sequence
                # to format the text, see http://en.wikipedia.org/wiki/ANSI_escape_code
                # A full escape sequence consists of the start elem '\x1b' and the
                # remaining string elem, which might be incomplete in this fragment.
                if i + 1 < lenSplit:
                    textAndNext = s + splitText[i + 1]
                    format, matchLen = self.parseFormat(
                        textAndNext, currentFormat, defaultFormat
                    )
                    if format is not None:
                        # escape sequence is complete and correct
                        splitText[i + 1] = splitText[i + 1][matchLen - 1 :]
                        ll.append(format)
                        currentFormat = format
                    else:
                        # escape sequence is incomplete or invalid
                        if i == lenSplit - 2:
                            # check if the split text ends with an incomplete format
                            for s2 in ("m", "0m", "[0m"):
                                if self._reFormatPattern.match(textAndNext + s2):
                                    unfinishedTailNew = textAndNext
                                    break
                        ll.append(s)
                        posInBlock += len(s)
                elif i == lenSplit - 1:
                    # escape sequence is incomplete or invalid
                    unfinishedTailNew = s
                    ll.append(s)
                    posInBlock += len(s)
            else:
                ll.append(s)
                posInBlock += len(s)

        self._unfinishedTail = unfinishedTailNew
        ll.append(posInBlock)
        finishedLines.append(ll)
        # the last line in finishedLines does not end with "\n"

        numCharsToRemove = cursor.positionInBlock() - lowestPosInFirstBlock

        # If necessary, make a new cursor that moves along. We insert
        # the text in pieces, so we need to move along with the text!
        if cursor.keepPositionOnInsert():
            cursor = QtGui.QTextCursor(cursor)
            cursor.setKeepPositionOnInsert(False)

        # delete characters from the previous fragment
        cursor.beginEditBlock()
        if numCharsToRemove > 0:
            cursor.movePosition(cursor.MoveOperation.Left, A_KEEP, numCharsToRemove)
            cursor.removeSelectedText()

        # shorten very long lines
        finishedLines = self._shortenLines(finishedLines, cursor)

        # write the text with proper formatting to the shell widget
        format = self._currentFormat
        for line in finishedLines:
            # lastPosInBlock = line[-1]
            for elem in line[:-1]:
                if isinstance(elem, str):
                    cursor.insertText(elem, format)
                elif isinstance(elem, QtGui.QTextCharFormat):
                    format = elem
        cursor.endEditBlock()

        self._currentFormat = currentFormat
        newLeftLimit = leftLimit
        return newLeftLimit

    @staticmethod
    def _splitStringIntoChunks(s, offset, chunkSize):
        n = len(s)
        numChunks, rem = divmod(n - offset, chunkSize)
        tillOffset = s[:offset]
        chunkList = [
            s[offset + i * chunkSize : offset + (i + 1) * chunkSize]
            for i in range(numChunks)
        ]
        remnant = s[offset + numChunks * chunkSize :]
        return tillOffset, chunkList, remnant

    def _shortenLines(self, finishedLines, cursor):
        SLM_NL = BaseShell.SHORTENED_LINE_MARKER + "\n"
        posInBlock = cursor.positionInBlock()
        finishedLinesShortened = []
        shortLineLength = 80  # length is without the newline char
        for indLine, line in enumerate(finishedLines):
            lastPosInBlock = line[-1]

            # hysteresis for line shortening
            if lastPosInBlock > 1024:
                self._lineShorteningActive = True
            elif lastPosInBlock <= shortLineLength + 1:  # +1 means including newline
                if indLine < len(finishedLines) - 1:
                    self._lineShorteningActive = False

            if self._lineShorteningActive:
                # Given a text, split the text into lines. Lines that are extremely
                # long are split in pieces of 80 characters to increase performance for
                # wrapping. This is kind of a failsafe for when the user accidentally
                # prints a bitmap or huge list. See https://github.com/pyzo/pyzo/issues/98

                if indLine == 0:
                    if posInBlock > shortLineLength:
                        # we also shorten the existing line
                        cursor.movePosition(
                            cursor.MoveOperation.Left, A_MOVE, posInBlock
                        )
                        multiples, rem = divmod(posInBlock, shortLineLength)
                        for i in range(multiples):
                            cursor.movePosition(
                                cursor.MoveOperation.Right, A_MOVE, shortLineLength
                            )
                            cursor.insertText(SLM_NL)
                        cursor.movePosition(cursor.MoveOperation.Right, A_MOVE, rem)
                        posInBlock = rem
                else:
                    posInBlock = 0

                lineNew = []
                for elem in line[:-1]:
                    if isinstance(elem, str):
                        n = len(elem)
                        if posInBlock + n > shortLineLength:
                            i = shortLineLength - posInBlock
                            tillOffset, chunkList, remnant = (
                                self._splitStringIntoChunks(elem, i, shortLineLength)
                            )
                            lineNew.append(tillOffset)
                            lineNew.append(SLM_NL)
                            lineNew.append(posInBlock + len(tillOffset) + len(SLM_NL))
                            posInBlock = 0
                            finishedLinesShortened.append(lineNew)

                            for chunk in chunkList:
                                finishedLinesShortened.append(
                                    [chunk, SLM_NL, shortLineLength + len(SLM_NL)]
                                )

                            lineNew = [remnant]
                            posInBlock = len(remnant)
                        else:
                            lineNew.append(elem)
                            posInBlock += n
                    else:
                        lineNew.append(elem)  # this elem is a QTextCharFormat
                lineNew.append(posInBlock)
                finishedLinesShortened.append(lineNew)
            else:
                finishedLinesShortened.append(line)

        return finishedLinesShortened

    def parseFormat(self, text, currentFormat, defaultFormat):
        """process ANSI escape codes for text formatting

        We only support a small subset of:
        http://en.wikipedia.org/wiki/ANSI_escape_code
        """
        mo = self._reFormatPattern.match(text)
        if not mo:
            return None, None
        format = QtGui.QTextCharFormat(currentFormat)
        params = [int(i) for i in mo[2].split(";")]
        matchLen = len(mo[1])
        for param in params:
            if param == 0:
                format = QtGui.QTextCharFormat(defaultFormat)
            elif param == 1:
                format.setFontWeight(QtGui.QFont.Weight.Bold)
            elif param == 2:
                format.setFontWeight(QtGui.QFont.Weight.Light)
            elif param == 3:
                format.setFontItalic(True)  # italic
            elif param == 4:
                format.setFontUnderline(True)  # underline
            elif param == 22:
                format.setFontWeight(QtGui.QFont.Weight.Normal)  # not bold or light
            elif param == 23:
                format.setFontItalic(False)  # not italic
            elif param == 24:
                format.setFontUnderline(False)  # not underline
            elif 30 <= param <= 37:  # set foreground color
                clr = self.COLORS[param - 30]
                format.setForeground(QtGui.QColor(clr))
            elif param == 39:  # reset the foreground color
                fg = defaultFormat.foreground()
                # we also need to set the brush style because it could be BrushStyle.NoBrush
                format.setForeground(QtGui.QBrush(fg.color(), fg.style()))
            elif 40 <= param <= 47:
                pass  # cannot set background text in QPlainTextEdit
            else:
                pass  # not supported
        return format, matchLen


class PythonShell(BaseShell):
    """The PythonShell class implements the python part of the shell
    by connecting to a remote process that runs a Python interpreter.
    """

    # Emits when the status string has changed or when receiving a new prompt
    stateChanged = QtCore.Signal(BaseShell)

    # Emits when the debug status is changed
    debugStateChanged = QtCore.Signal(BaseShell)

    def __init__(self, parent, info):
        super().__init__(parent)

        # Get standard info if not given.
        if info is None and pyzo.config.shellConfigs2:
            info = pyzo.config.shellConfigs2[0]
        if not info:
            info = KernelInfo(None)

        # Store info so we can reuse it on a restart
        self._info = info

        # For the editor to keep track of attempted imports
        self._importAttempts = []

        # To keep track of the response for introspection
        self._currentCTO = None
        self._currentACO = None

        # Write buffer to store messages in for writing
        self._write_buffer = None

        # Create timer to keep polling any results
        # todo: Maybe use yoton events to process messages as they arrive.
        # I tried this briefly, but it seemd to be less efficient because
        # messages are not so much bach-processed anymore. We should decide
        # on either method.
        self._timer = QtCore.QTimer(self)
        self._timer.setInterval(POLL_TIMER_INTERVAL_IDLE)  # ms
        self._timer.setSingleShot(False)
        self._timer.timeout.connect(self.poll)
        self._timer.start()

        # Add context menu
        self._menu = ShellContextMenu(shell=self, parent=self)
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(
            lambda p: self._menu.popup(self.mapToGlobal(p + QtCore.QPoint(0, 3)))
        )

        # Keep track of breakpoints
        pyzo.editors.breakPointsChanged.connect(self.sendBreakPoints)

        # Start!
        self.resetVariables()
        self.connectToKernel(info)

    def resetVariables(self):
        """Resets some variables."""

        # Reset read state
        self.setReadOnly(False)

        # Variables to store state, python version, builtins and keywords
        self._state = ""
        self._debugState = {}
        self._version = ""
        self._builtins = []
        self._keywords = []
        self._softKeywords = []
        self._startup_info = {}
        self._start_time = 0

        # (re)set import attempts
        self._importAttempts[:] = []

        # Update
        self.stateChanged.emit(self)

    def connectToKernel(self, info):
        """Create kernel and connect to it."""

        # Create yoton context
        self._context = ct = yoton.Context()

        # Create stream channels
        self._strm_out = yoton.SubChannel(ct, "strm-out")
        self._strm_err = yoton.SubChannel(ct, "strm-err")
        self._strm_raw = yoton.SubChannel(ct, "strm-raw")
        self._strm_echo = yoton.SubChannel(ct, "strm-echo")
        self._strm_prompt = yoton.SubChannel(ct, "strm-prompt")
        self._strm_broker = yoton.SubChannel(ct, "strm-broker")
        self._strm_action = yoton.SubChannel(ct, "strm-action", yoton.OBJECT)

        # Set channels to sync mode. This means that if Pyzo cannot process
        # the messages fast enough, the sending side is blocked for a short
        # while. We don't want our users to miss any messages.
        for c in [self._strm_out, self._strm_err]:
            c.set_sync_mode(True)

        # Create control channels
        self._ctrl_command = yoton.PubChannel(ct, "ctrl-command")
        self._ctrl_code = yoton.PubChannel(ct, "ctrl-code", yoton.OBJECT)
        self._ctrl_broker = yoton.PubChannel(ct, "ctrl-broker")

        # Create status channels
        self._stat_interpreter = yoton.StateChannel(ct, "stat-interpreter")
        self._stat_cd = yoton.StateChannel(ct, "stat-cd")
        self._stat_debug = yoton.StateChannel(ct, "stat-debug", yoton.OBJECT)
        self._stat_startup = yoton.StateChannel(ct, "stat-startup", yoton.OBJECT)
        self._stat_breakpoints = yoton.StateChannel(
            ct, "stat-breakpoints", yoton.OBJECT
        )

        self._stat_startup.received.bind(self._onReceivedStartupInfo)

        # Create introspection request channel
        default_future_timeout = 0.5  # default timeout for returned Future objects
        self._request = yoton.ReqChannel(ct, "reqp-introspect", default_future_timeout)

        # Connect! The broker will only start the kernel AFTER
        # we connect, so we do not miss out on anything.
        slot = pyzo.localKernelManager.createKernel(finishKernelInfo(info))
        self._brokerConnection = ct.connect("localhost:{}".format(slot))
        self._brokerConnection.closed.bind(self._onConnectionClose)

        # Force updating of breakpoints
        pyzo.editors.updateBreakPoints()

        # todo: see polling vs events

    #         # Detect incoming messages
    #         for c in [self._strm_out, self._strm_err, self._strm_raw,
    #                 self._strm_echo, self._strm_prompt, self._strm_broker,
    #                 self._strm_action,
    #                 self._stat_interpreter, self._stat_debug]:
    #             c.received.bind(self.poll)

    def get_kernel_cd(self):
        """Get current working dir of kernel.

        Returns the working directory, or "?" if it could not be determined.
        """
        return self._stat_cd.recv()

    def _onReceivedStartupInfo(self, channel):
        startup_info = channel.recv()

        # Store the whole dict
        self._startup_info = startup_info

        # Store when we received this
        self._start_time = time.time()

        # Set version
        version = startup_info.get("version", None)
        if isinstance(version, tuple):
            if version < (3,):
                self.setParser("python2")
            else:
                self.setParser("python3")
            self._version = "{}.{}".format(*version[:2])

        # Set keywords
        L = startup_info.get("keywords", None)
        if isinstance(L, list):
            self._keywords = L

        # Set soft keywords
        L = startup_info.get("softkeywords", None)
        if isinstance(L, list):
            self._softKeywords = L

        # Set builtins
        L = startup_info.get("builtins", None)
        if isinstance(L, list):
            self._builtins = L

        # Notify
        self.stateChanged.emit(self)

    ## Introspection processing methods

    def processCallTip(self, cto):
        """Processes a calltip request using a CallTipObject instance."""

        # Try using buffer first (not if we're not the requester)
        if self is cto.textCtrl:
            if cto.tryUsingBuffer():
                return

        # Clear buffer to prevent doing a second request
        # and store cto to see whether the response is still wanted.
        cto.setBuffer("")
        self._currentCTO = cto

        # Post request
        if cto.useIntermediateResult:
            future = self._request.signatureWithIntermediateResult(cto.name)
        else:
            future = self._request.signature(cto.name)
        future.add_done_callback(self._processCallTip_response)
        future.cto = cto

    def _processCallTip_response(self, future):
        """Process response of shell to show signature."""

        # Process future
        if future.cancelled():
            # print('Introspect cancelled')  # No kernel
            return
        elif future.exception():
            print("Introspect-exception: ", future.exception())
            return
        else:
            response = future.result()
            cto = future.cto

        # First see if this is still the right editor (can also be a shell)
        editor1 = pyzo.editors.getCurrentEditor()
        editor2 = pyzo.shells.getCurrentShell()
        if cto.textCtrl not in [editor1, editor2]:
            # The editor or shell starting the autocomp is no longer active
            cto.textCtrl.autocompleteCancel()
            return

        # Invalid response
        if not response:
            cto.textCtrl.autocompleteCancel()
            return

        if cto.useIntermediateResult:
            assert response.startswith("__pyzo__calltip")
            response = cto.name + response[len("__pyzo__calltip") :]
        # If still required, show tip, otherwise only store result
        if cto is self._currentCTO:
            cto.finish(response)
        else:
            cto.setBuffer(response)

    def processAutoComp(self, aco):
        """Processes an autocomp request using an AutoCompObject instance."""

        # Try using buffer first (not if we're not the requester)
        if self is aco.textCtrl:
            if aco.tryUsingBuffer():
                return

        # Include builtins and keywords?
        if not aco.name:
            aco.addNames(self._builtins)
            if pyzo.config.settings.autoComplete_keywords:
                aco.addNames(self._keywords)
                aco.addNames(self._softKeywords)
        elif aco.name == "[":
            return

        # Set buffer to prevent doing a second request
        # and store aco to see whether the response is still wanted.
        aco.setBuffer()
        self._currentACO = aco

        # Post request
        if aco.useIntermediateResult:
            if aco.name.endswith("["):  # if key auto-completion
                future = self._request.dir2WithIntermediateResult(aco.name[:-1])
            else:
                future = self._request.dirWithIntermediateResult(aco.name)
        else:
            if aco.name.endswith("["):  # if attribute or name auto-completion
                future = self._request.dir2(aco.name[:-1])
            else:
                future = self._request.dir(aco.name)
        future.add_done_callback(self._processAutoComp_response)
        future.aco = aco

    def _processAutoComp_response(self, future):
        """Process the response of the shell for the auto completion."""

        # Process future
        if future.cancelled():
            # print('Introspect cancelled')  # No living kernel
            return
        elif future.exception():
            print("Introspect-exception: ", future.exception())
            return
        else:
            response = future.result()
            aco = future.aco

        # First see if this is still the right editor (can also be a shell)
        editor1 = pyzo.editors.getCurrentEditor()
        editor2 = pyzo.shells.getCurrentShell()
        if aco.textCtrl not in [editor1, editor2]:
            # The editor or shell starting the autocomp is no longer active
            aco.textCtrl.autocompleteCancel()
            return

        # Add result to the list
        foundNames = []
        if response is not None:
            if aco.name.endswith("["):  # if key auto-completion
                foundNames = [
                    name[1:-1]
                    for name, type_, kind, repr_ in response
                    if name[0] == "["
                ]
                if response and response[-1][0] == "# ... too many entries":
                    foundNames.append(response[-1][0])
            else:  # if attribute or name auto-completion
                foundNames = response
        aco.addNames(foundNames)

        # Process list
        if aco.name and not foundNames and not aco.names:
            # No names found for the requested name, and no names from
            # fictive class members (via code parser in the editor).
            # Let's try to import it.
            importNames, importLines = pyzo.parser.getFictiveImports(editor1)
            baseName = aco.nameInImportNames(importNames)
            if baseName:
                line = importLines[baseName].strip()
                if line not in self._importAttempts:
                    # Do import
                    self.processLine(line + "  # auto-import")
                    self._importAttempts.append(line)
                    # Wait a barely noticable time to increase the chances
                    # that the import is complete when we repost the request.
                    time.sleep(0.2)
                    # To be sure, decrease the expiration date on the buffer
                    aco.setBuffer(timeout=1)

                    # repost request; self._importAttempts will prevent infinite loop
                    self.processAutoComp(aco)
        else:
            # If still required, show list, otherwise only store result
            if self._currentACO is aco:
                aco.finish()
            else:
                aco.setBuffer()

    ## Methods for executing code

    def executeCommand(self, text):
        """Execute one-line command in the remote Python session."""

        # Ensure edit line is selected (to reset scrolling to end)
        self.ensureCursorAtEditLine()

        self._ctrl_command.send(text)

    def executeCode(self, text, fname, lineno=None, cellName=None, changeDir=False):
        """Execute (run) a large piece of code in the remote shell.

        text: the source code to execute
        filename: the file from which the source comes
        lineno: the first lineno of the text in the file, where 0 would be
        the first line of the file...

        The text (source code) is first pre-processed:
        - convert all line-endings to \n
        - remove all empty lines at the end
        - remove commented lines at the end
        - convert tabs to spaces
        - dedent so minimal indentation is zero
        """

        # Convert tabs to spaces
        text = text.replace("\t", " " * 4)

        # Make sure there is always *some* text
        if not text:
            text = " "

        if lineno is None:
            lineno = 0
            cellName = fname  # run all

        # Examine the text line by line...
        # - check for empty/commented lined at the end
        # - calculate minimal indentation
        lines = text.splitlines()
        lastLineOfCode = 0
        minIndent = 99
        for linenr, line in enumerate(lines):
            # Check if empty (can be commented, but nothing more)
            tmp = line.split("#", 1)[0]  # get part before first #
            if tmp.count(" ") == len(tmp):
                continue  # empty line, proceed
            else:
                lastLineOfCode = linenr
            # Calculate indentation
            tmp = line.lstrip(" ")
            indent = len(line) - len(tmp)
            if indent < minIndent:
                minIndent = indent

        # Copy all proper lines to a new list,
        # remove minimal indentation, but only if we then would only remove
        # spaces (in the case of commented lines)
        lines2 = []
        for linenr in range(lastLineOfCode + 1):
            line = lines[linenr]
            # Remove indentation,
            if line[:minIndent].count(" ") == minIndent:
                line = line[minIndent:]
            else:
                line = line.lstrip(" ")
            lines2.append(line)

        # Ensure edit line is selected (to reset scrolling to end)
        self.ensureCursorAtEditLine()

        # Send message
        text = "\n".join(lines2)
        msg = {
            "source": text,
            "fname": fname,
            "lineno": lineno,
            "cellName": cellName,
            "changeDir": int(changeDir),
        }
        self._ctrl_code.send(msg)

    def sendBreakPoints(self, breaks):
        """Send all breakpoints."""
        # breaks is a dict of filenames to integers
        self._stat_breakpoints.send(breaks)

    ## The polling methods and terminating methods

    def poll(self, channel=None):
        """To keep the shell up-to-date. Call this periodically."""
        idle = True

        if self._write_buffer:
            # There is still data in the buffer
            sub, M = self._write_buffer
        else:
            # Check what subchannel has the latest message pending
            sub = yoton.select_sub_channel(
                self._strm_out,
                self._strm_err,
                self._strm_echo,
                self._strm_raw,
                self._strm_broker,
                self._strm_prompt,
            )
            # Read messages from it
            if sub:
                M = sub.recv_selected()
                # M = [sub.recv()] # Slow version (for testing)

        # Write all pending messages that are later than any other message
        if sub:
            idle = False
            # Select messages to process
            N = 256
            M, buffer = M[:N], M[N:]
            # Buffer the rest
            if buffer:
                self._write_buffer = sub, buffer
            else:
                self._write_buffer = None
            # Get how to deal with prompt
            prompt = 0
            if sub is self._strm_echo:
                prompt = 1
            elif sub is self._strm_prompt:
                prompt = 2
                M = M[-1:]  # only use the newest prompt
            # Get color
            color = None
            if sub is self._strm_broker:
                color = "#fff" if pyzo.darkSyntax else "#000"
            elif sub is self._strm_raw:
                color = "#bbb" if pyzo.darkSyntax else "#888888"
            elif sub is self._strm_err:
                color = "#f00"
            # Write
            self.write("".join(M), prompt, color, sub)

        # Do any actions?
        action = self._strm_action.recv(False)
        if action:
            idle = False
            if action == "cls":
                self.clearScreen()
            elif action.startswith("open "):
                parts = action.split(" ")
                parts.pop(0)
                try:
                    linenr = int(parts[0])
                    parts.pop(0)
                except ValueError:
                    linenr = None
                fname = " ".join(parts)
                editor = pyzo.editors.loadFile(fname)
                if editor and linenr:
                    editor._editor.gotoLine(linenr)
            else:
                print("Unknown action: {}".format(action))

        # ----- status

        newInterval = POLL_TIMER_INTERVAL_IDLE if idle else POLL_TIMER_INTERVAL_BUSY
        if self._timer.interval() != newInterval:
            self._timer.setInterval(newInterval)

        # Do not update status when the kernel is not really up and running
        # self._version is set when the startup info is received
        if not self._version:
            return

        # Update status
        state = self._stat_interpreter.recv()
        if state != self._state:
            self._state = state
            self.stateChanged.emit(self)

        # Update debug status
        state = self._stat_debug.recv()
        if state != self._debugState:
            self._debugState = state
            self.debugStateChanged.emit(self)

    def interrupt(self):
        """Send a Keyboard interrupt signal to the main thread of the remote process."""
        # Ensure edit line is selected (to reset scrolling to end)
        self.ensureCursorAtEditLine()

        self._ctrl_broker.send("INT")

    def pause(self):
        """Send a pause signal to the main thread of the remote process."""
        # Ensure edit line is selected (to reset scrolling to end)
        self.ensureCursorAtEditLine()

        self._ctrl_broker.send("PAUSE")

    def restart(self, scriptFile=None):
        """Terminate the shell, after which it is restarted.

        Args can be a filename, to execute as a script as soon as the
        shell is back up.
        """
        self.resetShellWriters()

        # Ensure edit line is selected (to reset scrolling to end)
        self.ensureCursorAtEditLine()

        # Get info
        info = finishKernelInfo(self._info, scriptFile)

        # Create message and send
        msg = "RESTART\n" + ssdf.saves(info)
        self._ctrl_broker.send(msg)

        # Reset
        self.resetVariables()

    def terminate(self):
        """Terminates the python process.

        It will first try gently, but if that does not work, the process
        shall be killed. To be notified of the termination, connect to the
        "terminated" signal of the shell.
        """
        # Ensure edit line is selected (to reset scrolling to end)
        self.ensureCursorAtEditLine()

        self._ctrl_broker.send("TERM")

    def closeShell(self):  # do not call it close(); that is a reserved method.
        """This closes the shell.

        If possible, we will first tell the broker to terminate the kernel.

        The broker will be cleaned up if there are no clients connected
        and if there is no active kernel. In a multi-user environment,
        we should thus be able to close the shell without killing the
        kernel. But in a closed 1-to-1 environment we really want to
        prevent loose brokers and kernels dangling around.

        In both cases however, it is the responsibility of the broker to
        terminate the kernel, and the shell will simply assume that this
        will work :)
        """

        pyzo.editors.breakPointsChanged.disconnect(self.sendBreakPoints)

        # If we can, try to tell the broker to terminate the kernel
        if self._context and self._context.connection_count:
            self.terminate()
            self._context.flush()  # Important, make sure the message is sent!
            self._context.close()

        # Adios
        pyzo.shells.removeShell(self)

        # This object (self) still stays in memory after closing the shell (memory leak).
        # At least, we try to clean up as much as possible.
        self._timer.stop()
        self._timer.timeout.disconnect()
        self.customContextMenuRequested.disconnect()

    def _onConnectionClose(self, c, why):
        """To be called after disconnecting.

        In general, the broker will not close the connection, so it can
        be considered an error-state if this function is called.
        """

        # Stop context
        if self._context:
            self._context.close()

        # New (empty prompt)
        self._cursor1.movePosition(self._cursor1.MoveOperation.End, A_MOVE)
        self._cursor2.movePosition(self._cursor2.MoveOperation.End, A_MOVE)

        self.write("\n\n")
        self.write("Lost connection with broker:\n")
        self.write(why)
        self.write("\n\n")

        # Set style to indicate dead-ness
        self.setReadOnly(True)

        # Goto end such that the closing message is visible
        cursor = self.textCursor()
        cursor.movePosition(cursor.MoveOperation.End, A_MOVE)
        self.setTextCursor(cursor)
        self.ensureCursorVisible()


if __name__ == "__main__":
    b = BaseShell(None)
    b.show()



import os

from pyzo.qt import QtCore, QtGui, QtWidgets

import pyzo
from pyzo.util import zon as ssdf
from pyzo.core.pyzoLogging import print
import pyzo.core.baseTextCtrl
from pyzo.codeeditor.style import StyleFormat


SAMPLE = """
## Foo class
# This is a comment
class Foo:
''' This class does nothing. '''
    #TODO: be amazing
        def baz(self, arg1):
            return max(arg1, 42)
    bar = "Hello wor
""" + chr(160)


class FakeEditor(pyzo.core.baseTextCtrl.BaseTextCtrl):
    """This "fake" editor emits a signal when
    the user clicks on a word with a token:
    a click on the word "class" emits with arg "syntax.keyword".

    It may be improved by adding text with specific token
    like Editor.text which are not present by default
    """

    tokenClicked = QtCore.Signal(str)

    def __init__(self, text=""):
        super().__init__()

        # set parser to enable syntaxic coloration
        self.setParser("python3")
        self.setReadOnly(False)
        self.setLongLineIndicatorPosition(30)
        self.setPlainText(SAMPLE)

    def mousePressEvent(self, event):
        super().mousePressEvent(event)

        # get the text position of the click
        pos = self.textCursor().columnNumber()
        tokens = self.textCursor().block().userData().tokens

        # Find the token which contains the click pos
        for tok in tokens:
            if tok.start <= pos <= tok.end:
                self.tokenClicked.emit(tok.description.key)
                break


class TitledWidget(QtWidgets.QWidget):
    """A litle helper class to "name" a widget :
    it displays a QLabel to left of the given widget"""

    def __init__(self, name, other):
        super().__init__()
        self.widget = other
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(QtWidgets.QLabel(text=name.capitalize().strip() + " :"))
        layout.addWidget(other)

        self.setLayout(layout)

    def setFocus(self):
        self.widget.setFocus()


class ColorLineEdit(QtWidgets.QLineEdit):
    """A subclass of the QLineEdit that can open
    a QColorDialog on click of a button
    """

    def __init__(self, name, *args, **kwargs):
        """The name is displayed in the QColorDialog"""
        super().__init__(*args, **kwargs)
        self.name = name
        self.button = QtWidgets.QToolButton(self)
        self.button.setIcon(QtGui.QIcon(pyzo.icons.cog))
        self.button.setStyleSheet("border: 0px; padding: 0px")
        self.button.clicked.connect(self.openColorDialog)

        frameWidth = self.style().pixelMetric(
            QtWidgets.QStyle.PixelMetric.PM_DefaultFrameWidth
        )
        buttonSize = self.button.sizeHint()

        self.setStyleSheet(
            "QLineEdit {{padding-right: {}px; }}".format(
                buttonSize.width() + frameWidth + 1
            )
        )
        # self.setMinimumSize(max(100, buttonSize.width() + frameWidth*2 + 2),
        #                     max(self.minimumSizeHint().height(), buttonSize.height() + frameWidth*2 + 2))

    def openColorDialog(self):
        """A simple function that opens a QColorDialog
        and link the dialog current color selection
        to the QLineEdit text
        """
        dlg = QtWidgets.QColorDialog(self)
        dlg.setWindowTitle("Pick a color for the " + self.name.lower())
        dlg.setCurrentColor(QtGui.QColor(self.text()))
        dlg.currentColorChanged.connect(lambda clr: self.setText(clr.name()))
        dlg.setModal(False)
        dlg.exec()

    def resizeEvent(self, event):
        buttonSize = self.button.sizeHint()
        frameWidth = self.style().pixelMetric(
            QtWidgets.QStyle.PixelMetric.PM_DefaultFrameWidth
        )
        self.button.move(
            int(self.rect().right() - frameWidth - buttonSize.width()),
            int(self.rect().bottom() - buttonSize.height() + 1) // 2,
        )
        super().resizeEvent(event)


class StyleEdit(QtWidgets.QWidget):
    """The StyleLineEdit is a line that allows the edition
    of one style (i.e. "Editor.Text" or  "Syntax.identifier")
    with a given StyleElementDescription it find the editable
    parts and display the adaptated widgets for edition
    (checkbok for bold and italic, combo box for linestyles...).
    """

    styleChanged = QtCore.Signal(str, str)

    def __init__(self, defaultStyle, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # The styleKey is sent with the styleChanged signal for easy identification
        self.styleKey = defaultStyle.key

        self.layout = layout = QtWidgets.QHBoxLayout()
        # The setters are used when setting the style
        self.setters = {}

        # TODO: the use of StyleFormat._parts should be avoided
        # We use the StyleFormat._parts keys, to find the elements
        # Useful to edits, because the property may return a value
        # Even if they were not defined in the defaultFormat
        fmtParts = defaultStyle.defaultFormat._parts

        # Add the widgets corresponding to the fields
        if "fore" in fmtParts:
            self.__add_clrLineEdit("fore", "Foreground")
        if "back" in fmtParts:
            self.__add_clrLineEdit("back", "Background")
        if "bold" in fmtParts:
            self.__add_checkBox("bold", "Bold")
        if "italic" in fmtParts:
            self.__add_checkBox("italic", "Italic")
        if "underline" in fmtParts:
            self.__add_comboBox(
                "underline", "Underline", "No", "Dotted", "Wave", "Full", "Yes"
            )
        if "linestyle" in fmtParts:
            self.__add_comboBox("linestyle", "Linestyle", "Dashed", "Dotted", "Full")

        self.setLayout(layout)

        SP = QtWidgets.QSizePolicy.Policy
        self.setSizePolicy(SP.Minimum, SP.Minimum)

    def __add_clrLineEdit(self, key, name):
        """this is a helper method to create a ColorLineEdit
        it adds the created widget (as a TitledWidget) to the layout and
        register a setter and listen to changes
        """
        clrEdit = ColorLineEdit(name)
        clrEdit.textChanged.connect(lambda txt, key=key: self.__update(key, txt))
        self.setters[key] = clrEdit.setText
        self.layout.addWidget(TitledWidget(name, clrEdit), 0)

    def __add_checkBox(self, key, name):
        """this is a helper method to create a QCheckBox
        it adds the created widget (as a TitledWidget) to the layout and
        register a setter and listen to changes
        """

        checkBox = QtWidgets.QCheckBox()
        self.setters[key] = lambda val, check=checkBox: check.setCheckState(
            [QtCore.Qt.CheckState.Unchecked, QtCore.Qt.CheckState.Checked][val == "yes"]
        )

        checkBox.stateChanged.connect(
            lambda state, key=key: self.__update(key, "yes" if state else "no")
        )

        self.layout.addWidget(TitledWidget(name, checkBox))

    def __add_comboBox(self, key, name, *items):
        """this is a helper method to create a comboBox
        it adds the created widget (as a TitledWidget) to the layout and
        register a setter and listen to changes
        """

        combo = QtWidgets.QComboBox()
        combo.addItems(items)
        combo.currentTextChanged.connect(lambda txt, key=key: self.__update(key, txt))

        # Note: those setters may become problematic if
        # someone use the synonyms (defined in codeeditor/style.py)
        # i.e. a stylement is of form "linestyle:dashline"
        # instead of the "linestyle:dashed"
        self.setters[key] = lambda txt, cmb=combo: cmb.setCurrentText(txt.capitalize())
        self.layout.addWidget(TitledWidget(name, combo))

    def __update(self, key, value):
        """this function is called everytime one of the children
        widget data has been modified by the user"""
        self.styleChanged.emit(self.styleKey, key + ":" + value)

    def setStyle(self, text):
        """updates every children to match the StyleFormat(text) fields"""
        style = StyleFormat(text)
        for key, setter in self.setters.items():
            setter(style[key])

    def setFocus(self):
        self.layout.itemAt(0).widget().setFocus()


class ThemeEditorWidget(QtWidgets.QWidget):
    """The ThemeEditorWidgets allows to edits themes,
    it has one StyleEdit widget per StyleElements ("Editor.Text",
    "Syntax.string"). It emits a signal on each style changes

    It also manages basic theme I/O :
        - adding new theme
        - renaming theme

    """

    styleChanged = QtCore.Signal(dict)
    done = QtCore.Signal(int)

    def __init__(self, themes, *args, editor=None, **kwargs):
        super().__init__(*args, **kwargs)

        # dict of themes, a deep copy of pyzo.themes
        self.themes = themes
        # We store the key name separate so we can easier track renames
        self.cur_theme_key = ""
        # The current theme being changed
        self.cur_theme = None

        # If an editor is given, connect to it
        self.editor = editor
        if self.editor is not None:
            self.editor.tokenClicked.connect(self.focusOnStyle)
            self.styleChanged.connect(self.editor.setStyle)

        # Display editables style formats in a scroll area
        self.scrollArea = scrollArea = QtWidgets.QScrollArea()
        self.scrollArea.setWidgetResizable(True)

        formLayout = QtWidgets.QFormLayout()
        self.styleEdits = {}

        # Add one pair of label and StyleEdit per style element description
        # to the formLayout and connect the StyleEdit signals to the updatedStyle method
        for styleDesc in pyzo.codeeditor.CodeEditor.getStyleElementDescriptions():
            label = QtWidgets.QLabel(text=styleDesc.name, toolTip=styleDesc.description)
            label.setWordWrap(True)
            styleEdit = StyleEdit(styleDesc, toolTip=styleDesc.description)
            styleEdit.styleChanged.connect(self.updatedStyle)
            self.styleEdits[styleDesc.key] = styleEdit
            formLayout.addRow(label, styleEdit)

        wrapper = QtWidgets.QWidget()
        wrapper.setLayout(formLayout)
        wrapper.setMinimumWidth(650)
        scrollArea.setWidget(wrapper)

        # Basic theme I/O

        curThemeLbl = QtWidgets.QLabel(text="Themes:")

        self.curThemeCmb = curThemeCmb = QtWidgets.QComboBox()
        current_index = -1
        for i, themeName in enumerate(self.themes.keys()):
            # We store the themeName in data in case the user renames one
            curThemeCmb.addItem(themeName, userData=themeName)
            if themeName == pyzo.config.settings.theme.lower():
                current_index = i
        curThemeCmb.addItem("New...")

        loadLayout = QtWidgets.QHBoxLayout()
        loadLayout.addWidget(curThemeLbl)
        loadLayout.addWidget(curThemeCmb)

        self.saveBtn = saveBtn = QtWidgets.QPushButton(text="Save")
        saveBtn.clicked.connect(self.saveTheme)
        exitBtn = QtWidgets.QPushButton(text="Apply theme")
        exitBtn.clicked.connect(self.ok)

        exitLayout = QtWidgets.QHBoxLayout()
        exitLayout.addWidget(exitBtn)
        exitLayout.addWidget(saveBtn)

        # Packing it up
        mainLayout = QtWidgets.QVBoxLayout()
        mainLayout.addLayout(loadLayout)
        mainLayout.addWidget(scrollArea)
        mainLayout.addLayout(exitLayout)
        self.setLayout(mainLayout)

        curThemeCmb.currentIndexChanged.connect(self.indexChanged)
        curThemeCmb.currentTextChanged.connect(self.setTheme)

        # Init
        if current_index >= 0:
            curThemeCmb.setCurrentIndex(current_index)
            self.setTheme(pyzo.config.settings.theme)

    def createTheme(self):
        """Create a new theme based on the current
        theme selected.
        """

        index = self.curThemeCmb.currentIndex()
        if index != self.curThemeCmb.count() - 1:
            return self.curThemeCmb.setCurrentIndex(self.curThemeCmb.count() - 1)

        # Select a new name
        t = "new_theme_x"
        i = 1
        themeName = t.replace("x", str(i))
        while themeName in self.themes:
            i += 1
            themeName = t.replace("x", str(i))

        # Create new theme
        new_theme = {"name": themeName, "data": {}, "builtin": False}
        if self.cur_theme:
            new_theme["data"] = self.cur_theme["data"].copy()
        self.cur_theme_key = themeName
        self.cur_theme = new_theme
        self.themes[themeName] = new_theme

        self.curThemeCmb.setItemText(index, themeName)
        self.curThemeCmb.setItemData(index, themeName)

        self.curThemeCmb.setEditable(True)
        self.curThemeCmb.lineEdit().setCursorPosition(0)
        self.curThemeCmb.lineEdit().selectAll()

        self.saveBtn.setEnabled(True)

        self.curThemeCmb.addItem(
            "New...",
        )

    def setTheme(self, name):
        """Set the theme by its name. The combobox becomes editable only
        if the theme is not builtin. This method is connected to the signal
        self.curThemeCmb.currentTextChanged ; so it also filters
        parasites events"""

        name = name.lower()

        if name != self.curThemeCmb.currentText():
            # An item was added to the comboBox
            # But it's not a user action so we quit
            print(" -> Cancelled because this was not a user action")
            return

        if self.cur_theme_key == self.curThemeCmb.currentData():
            # The user renamed an existing theme
            self.cur_theme["name"] = name
            return

        if name not in self.themes:
            return

        # Sets the curent theme key
        self.cur_theme_key = name
        self.cur_theme = self.themes[name]

        if self.cur_theme["builtin"]:
            self.saveBtn.setEnabled(False)
            self.saveBtn.setText("Cannot save builtin style")
        else:
            self.saveBtn.setEnabled(True)
            self.saveBtn.setText("Save")
        self.curThemeCmb.setEditable(not self.cur_theme["builtin"])

        for key, le in self.styleEdits.items():
            if key in self.cur_theme["data"]:
                try:
                    le.setStyle(self.cur_theme["data"][key])
                except Exception as e:
                    print(
                        "Exception while setting style", key, "for theme", name, ":", e
                    )

    def saveTheme(self):
        """Saves the current theme to the disk, in appDataDir/themes"""

        if self.cur_theme["builtin"]:
            return
        themeName = self.curThemeCmb.currentText().strip()
        if not themeName:
            return

        # Get user theme dir and make sure it exists
        dir = os.path.join(pyzo.appDataDir, "themes")
        os.makedirs(dir, exist_ok=True)

        # Try to delete the old file if it exists (useful if it was renamed)
        try:
            os.remove(os.path.join(dir, self.cur_theme_key + ".theme"))
        except Exception:
            pass

        # This is the needed because of the SSDF format:
        # it doesn't accept dots, so we put underscore instead
        data = {x.replace(".", "_"): y for x, y in self.cur_theme["data"].items()}

        fname = os.path.join(dir, themeName + ".theme")
        ssdf.save(fname, {"name": themeName, "data": data})
        print("Saved theme '{}' to '{}'".format(themeName, fname))

    def ok(self):
        """On user click saves the cur_theme if modified
        and restart pyzo if the theme changed"""
        prev = pyzo.config.settings.theme
        new = self.cur_theme["name"]

        self.saveTheme()

        if prev != new:
            pyzo.config.settings.theme = new
            pyzo.saveConfig()
            # This may be better
            pyzo.main.restart()
        else:
            self.done.emit(1)

    def indexChanged(self, index):
        # User selected the "new..." button
        if index == self.curThemeCmb.count() - 1:
            self.createTheme()

    def focusOnStyle(self, key):
        self.styleEdits[key].setFocus()
        self.scrollArea.ensureWidgetVisible(self.styleEdits[key])

    def updatedStyle(self, style, text):
        fmt = StyleFormat(self.cur_theme["data"][style])
        fmt.update(text)
        self.cur_theme["data"][style] = str(fmt)
        self.styleChanged.emit({style: text})


class EditColorDialog(QtWidgets.QDialog):
    """This dialog allows to edit color schemes,
    it is composed of two main components :
        - a "fake" editor to visualize the changes
        - a theme editor to make the edits
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle("Color scheme")

        # Make a deep copy
        themes = {}
        for name, theme in pyzo.themes.items():
            theme = theme.copy()
            theme["data"] = theme["data"].copy()
            themes[name] = theme

        self.editor = FakeEditor()
        self.editColor = ThemeEditorWidget(themes=themes, editor=self.editor)
        self.editColor.done.connect(self.done)
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self.editor, 1)
        layout.addWidget(self.editColor, 2)
        self.setLayout(layout)

    def sizeHint(self):
        # We use this instead of self.resize(1200, 800) to avoid
        # a too large dialog window on smaller screens.
        return QtCore.QSize(1200, 800)



from yoton.connection import Connection


class ItcConnection(Connection):
    """ItcConnection(context, hostname, port, name='')

    Not implemented .

    The inter-thread-communication connection class implements a
    connection between two contexts that are in the same process.
    Two instances of this class are connected using a weak reference.
    In case one of the ends is cleaned up by the garbadge collector,
    the other end will close the connection.

    """

    pass
    # todo: implement me



"""Module yoton.channels.message_types

Defines a few basic message_types for the channels. A MessageType object
defines how a message of that type should be converted to bytes and
vice versa.

The Packer and Unpacker classes for the ObjectMessageType are based on
the xdrrpc Python module by Rob Reilink and Windel Bouwman.

"""

import sys
import struct
from yoton.misc import bytes, basestring, long

# To decode P2k strings that are not unicode
if sys.__stdin__ and sys.__stdin__.encoding:
    STDINENC = sys.__stdin__.encoding
elif sys.stdin and sys.stdin.encoding:
    STDINENC = sys.stdin.encoding
else:
    STDINENC = "utf-8"


class MessageType(object):
    """MessageType()

    Instances of this class are used to convert messages to bytes and
    bytes to messages.

    Users can easily inherit from this class to make channels work for
    user specific message types. Three methods should be overloaded:
      * message_to_bytes()  - given a message, returns bytes
      * message_from_bytes() - given bytes, returns the message
      * message_type_name() - a string (for example 'text', 'array')

    The message_type_name() method is used by the channel to add an
    extension to the slot name, such that only channels of the same
    message type (well, with the same message type name) can connect.

    """

    def message_to_bytes(self, message):
        raise NotImplementedError()

    def message_from_bytes(self, bb):
        raise NotImplementedError()

    def message_type_name(self):
        raise NotImplementedError()


class BinaryMessageType(MessageType):
    """BinaryMessageType()

    To let channels handle binary messages.
    Available as yoton.BINARY.

    """

    def message_type_name(self):
        return "bin"

    def message_to_bytes(self, message):
        if not isinstance(message, bytes):
            raise ValueError("Binary channel requires byte messages.")
        return message

    def message_from_bytes(self, bb):
        return bb


class TextMessageType(MessageType):
    """BinaryMessageType()

    To let channels handle Unicode text messages.
    Available as yoton.TEXT.

    """

    def message_type_name(self):
        return "txt"

    def message_to_bytes(self, message):
        # Check
        if not isinstance(message, basestring):
            raise ValueError("Text channel requires string messages.")

        # If using py2k and the string is not unicode, make unicode first
        # by try encoding using UTF-8. When a piece of code stored
        # in a unicode string is executed, str objects are utf-8 encoded.
        # Otherwise they are encoded using __stdin__.encoding. In specific
        # cases, a non utf-8 encoded str might be succesfully encoded
        # using utf-8, but this is rare. Since I would not
        # know how to tell the encoding beforehand, we'll take our
        # chances... Note that in Pyzo (for which this package was created,
        # all executed code is unicode, so str instrances are always
        # utf-8 encoded.
        if isinstance(message, bytes):
            try:
                message = message.decode("utf-8")
            except UnicodeError:
                try:
                    message = message.decode(STDINENC)
                except UnicodeError:
                    # Probably not really a string then?
                    message = repr(message)

        # Encode and send
        return message.encode("utf-8")

    def message_from_bytes(self, bb):
        return bb.decode("utf-8")


class ObjectMessageType(MessageType):
    """ObjectMessageType()

    To let channels handle messages consisting of any of the following
    Python objects: None, bool, int, float, string, list, tuple, dict.
    Available as yoton.OBJECT.

    """

    def message_type_name(self):
        return "obj"

    def message_to_bytes(self, message):
        packer = Packer()
        packer.pack_object(message)
        return packer.get_buffer()

    def message_from_bytes(self, bb):
        if bb:
            unpacker = Unpacker(bb)
            return unpacker.unpack_object()
        else:
            return None


# Formats
_FMT_TYPE = "<B"
_FMT_BOOL = "<B"
_FMT_INT = "<q"
_FMT_FLOAT = "<d"

# Types
_TYPE_NONE = ord("n")
_TYPE_BOOL = ord("b")
_TYPE_INT = ord("i")
_TYPE_FLOAT = ord("f")
_TYPE_STRING = ord("s")
_TYPE_LIST = ord("l")
_TYPE_TUPLE = ord("t")
_TYPE_DICT = ord("d")


class Packer:
    # Note that while xdrlib uses StringIO/BytesIO, this approach using
    # a list is actually faster.

    def __init__(self):
        self._buf = []

    def get_buffer(self):
        return bytes().join(self._buf)

    def write(self, bb):
        self._buf.append(bb)

    def write_number(self, n):
        if n < 255:
            self.write(struct.pack("<B", n))
        else:
            self.write(struct.pack("<B", 255))
            self.write(struct.pack("<Q", n))

    def pack_object(self, object):
        if object is None:
            self.write(struct.pack(_FMT_TYPE, _TYPE_NONE))
        elif isinstance(object, bool):
            self.write(struct.pack(_FMT_TYPE, _TYPE_BOOL))
            self.write(struct.pack(_FMT_BOOL, object))
        elif isinstance(object, (int, long)):
            self.write(struct.pack(_FMT_TYPE, _TYPE_INT))
            self.write(struct.pack(_FMT_INT, object))
        elif isinstance(object, float):
            self.write(struct.pack(_FMT_TYPE, _TYPE_FLOAT))
            self.write(struct.pack(_FMT_FLOAT, object))
        elif isinstance(object, basestring):
            bb = object.encode("utf-8")
            self.write(struct.pack(_FMT_TYPE, _TYPE_STRING))
            self.write_number(len(bb))
            self.write(bb)
        elif isinstance(object, list):
            self.write(struct.pack(_FMT_TYPE, _TYPE_LIST))
            self.write_number(len(object))
            for value in object:
                self.pack_object(value)  # call recursive
        elif isinstance(object, tuple):
            self.write(struct.pack(_FMT_TYPE, _TYPE_TUPLE))
            self.write_number(len(object))
            for value in object:
                self.pack_object(value)  # call recursive
        elif isinstance(object, dict):
            self.write(struct.pack(_FMT_TYPE, _TYPE_DICT))
            self.write_number(len(object))
            # call recursive
            for key in object:
                self.pack_object(key)
                self.pack_object(object[key])
        else:
            raise ValueError("Unsupported type: %s" % repr(type(object)))


class Unpacker:
    def __init__(self, data):
        self._buf = data
        self._pos = 0

    def read(self, n):
        i1 = self._pos
        i2 = self._pos + n
        if i2 > len(self._buf):
            raise EOFError
        else:
            self._pos = i2
            return self._buf[i1:i2]

    def read_number(self):
        (n,) = struct.unpack("<B", self.read(1))
        if n == 255:
            (n,) = struct.unpack("<Q", self.read(8))
        return n

    def unpack(self, fmt, n):
        i1 = self._pos
        i2 = self._pos + n
        if i2 > len(self._buf):
            raise EOFError
        else:
            self._pos = i2
            data = self._buf[i1:i2]
            return struct.unpack(fmt, data)[0]

    def unpack_object(self):
        object_type = self.unpack(_FMT_TYPE, 1)

        if object_type == _TYPE_NONE:
            return None
        elif object_type == _TYPE_BOOL:
            return bool(self.unpack(_FMT_BOOL, 1))
        elif object_type == _TYPE_INT:
            return self.unpack(_FMT_INT, 8)
        elif object_type == _TYPE_FLOAT:
            return self.unpack(_FMT_FLOAT, 8)
        elif object_type == _TYPE_STRING:
            n = self.read_number()
            return self.read(n).decode("utf-8")
        elif object_type == _TYPE_LIST:
            object = [self.unpack_object() for i in range(self.read_number())]
            return object
        elif object_type == _TYPE_TUPLE:
            object = [self.unpack_object() for i in range(self.read_number())]
            return tuple(object)
        elif object_type == _TYPE_DICT:
            object = {}
            for i in range(self.read_number()):
                key = self.unpack_object()
                object[key] = self.unpack_object()
            return object
        else:
            raise ValueError("Unsupported type: %s" % repr(object_type))


# Define constants
TEXT = TextMessageType()
BINARY = BinaryMessageType()
OBJECT = ObjectMessageType()


if __name__ == "__main__":
    # Test

    s = {}
    s["foo"] = 3
    s["bar"] = 9
    s["empty"] = []
    s[(2, "aa", 3)] = ["pretty", ("nice", "eh"), 4]

    bb = OBJECT.message_to_bytes(s)
    s2 = OBJECT.message_from_bytes(bb)
    print(s)
    print(s2)



"""Module style

Provides basic functionality for styling.

Styling is done using a dictionary of StyleFormat instances. Each
such instance represents a certain element being styled (e.g. keywords,
line numbers, indentation guides).

All possible style elements are represented using StyleElementDescription
instances. These have a name, description and default format, which
makes it easy to build a UI to allow the user to change the style.

"""

from .qt import QtGui, QtCore

Qt = QtCore.Qt


class StyleElementDescription:
    """StyleElementDescription(name, defaultFormat, description)

    Describes a style element by its name, default format, and description.

    A style description is a simple placeholder for something
    that can be styled.

    """

    def __init__(self, name, description, defaultFormat):
        self._name = name
        self._description = description
        self._defaultFormat = StyleFormat(defaultFormat)

    def __repr__(self):
        return '<"{}": "{}">'.format(self.name, self.defaultFormat)

    @property
    def name(self):
        return self._name

    @property
    def key(self):
        return self._name.replace(" ", "").lower()

    @property
    def description(self):
        return self._description

    @property
    def defaultFormat(self):
        return self._defaultFormat


class StyleFormat:
    """StyleFormat(format='')

    Represents the style format for a specific style element.
    A "style" is a dictionary that maps names (of style elements)
    to StyleFormat instances.

    The given format can be a string or another StyleFormat instance.
    Style formats can be combined using their update() method.

    A style format consists of multiple parts, where each "part" consists
    of a key and a value. The keys can be anything, depending
    on what kind of thing is being styled. The value can be obtained using
    the index operator (e.g. styleFomat['fore'])

    For a few special keys, properties are defined that return the Qt object
    corresponding to the value. These values are also buffered to enable
    fast access. These keys are:
      * fore: (QColor) the foreground color
      * back: (QColor) the background color
      * bold: (bool) whether the text should be bold
      * italic: (bool) whether the text should be in italic
      * underline: (int) whether an underline should be used (and which one)
      * linestyle: (int) what line style to use (e.g. for indent guides)
      * textCharFormat: (QTextCharFormat) for the syntax styles

    The format neglects spaces and case. Parts are separated by commas
    or semicolons. If only a key is given it's value is interpreted
    as 'yes'. If only a color is given, its key is interpreted as 'fore'
    and back. Colors should be given using the '#' hex formatting.

    An example format string: 'fore:#334, bold, underline:dotLine'

    By calling str(styleFormatInstance) the string representing of the
    format can be obtained. By iterating over the instance, a series
    of key-value pairs is obtained.

    """

    def __init__(self, format=""):
        self._parts = {}
        self.update(format)

    def _resetProperties(self):
        self._fore = None
        self._back = None
        self._bold = None
        self._italic = None
        self._underline = None
        self._linestyle = None
        self._textCharFormat = None

    def __str__(self):
        """Get a (cleaned up) string representation of this style format."""
        parts = ["{}:{}".format(key, self._parts[key]) for key in self._parts]
        return ", ".join(parts)

    def __repr__(self):
        return '<StyleFormat "{}">'.format(self)

    def __getitem__(self, key):
        try:
            return self._parts[key]
        except KeyError:
            raise KeyError("Invalid part key " + key + " for style format.")

    def __iter__(self):
        """Yields a series of tuples (key, val)."""
        parts = [(key, self._parts[key]) for key in self._parts]
        return parts.__iter__()

    def update(self, format):
        """Update this style format with the given format."""

        # Reset buffered values
        self._resetProperties()

        # Make a string, so we update the format with the given one
        if isinstance(format, StyleFormat):
            format = str(format)

        # Split on ',' and ','
        styleParts = format.replace("=", ":").replace(";", ",").split(",")

        for stylePart in styleParts:
            # Make sure it consists of identifier and value pair
            # e.g. fore:#xxx, bold:yes, underline:no
            if ":" not in stylePart:
                if stylePart.startswith("#"):
                    stylePart = "foreandback:" + stylePart
                else:
                    stylePart += ":yes"

            # Get key value and strip and make lowecase
            key, _, val = [i.strip().lower() for i in stylePart.partition(":")]

            # Store in parts
            if key == "foreandback":
                self._parts["fore"] = val
                self._parts["back"] = val
            elif key:
                self._parts[key] = val

    ## Properties

    def _getValueSafe(self, key):
        try:
            return self._parts[key]
        except KeyError:
            return "no"

    @property
    def fore(self):
        if self._fore is None:
            self._fore = QtGui.QColor(self._parts["fore"])
        return self._fore

    @property
    def back(self):
        if self._back is None:
            self._back = QtGui.QColor(self._parts["back"])
        return self._back

    @property
    def bold(self):
        if self._bold is None:
            self._bold = self._getValueSafe("bold") in ["yes", "true"]
        return self._bold

    @property
    def italic(self):
        if self._italic is None:
            self._italic = self._getValueSafe("italic") in ["yes", "true"]
        return self._italic

    @property
    def underline(self):
        if self._underline is None:
            US = QtGui.QTextCharFormat.UnderlineStyle
            val = self._getValueSafe("underline")
            if val in ["yes", "true"]:
                self._underline = US.SingleUnderline
            elif val in ["dotted", "dots", "dotline"]:
                self._underline = US.DotLine
            elif val in ["wave"]:
                self._underline = US.WaveUnderline
            else:
                self._underline = US.NoUnderline
        return self._underline

    @property
    def linestyle(self):
        if self._linestyle is None:
            val = self._getValueSafe("linestyle")
            if val in ["yes", "true"]:
                self._linestyle = Qt.PenStyle.SolidLine
            elif val in ["dotted", "dot", "dots", "dotline"]:
                self._linestyle = Qt.PenStyle.DotLine
            elif val in ["dashed", "dash", "dashes", "dashline"]:
                self._linestyle = Qt.PenStyle.DashLine
            else:
                self._linestyle = Qt.PenStyle.SolidLine  # default to solid
        return self._linestyle

    @property
    def textCharFormat(self):
        if self._textCharFormat is None:
            self._textCharFormat = QtGui.QTextCharFormat()
            self._textCharFormat.setForeground(self.fore)
            try:  # not all styles have a back property
                self._textCharFormat.setBackground(self.back)
            except Exception:
                pass
            self._textCharFormat.setUnderlineStyle(self.underline)
            if self.bold:
                self._textCharFormat.setFontWeight(QtGui.QFont.Weight.Bold)
            if self.italic:
                self._textCharFormat.setFontItalic(True)
        return self._textCharFormat



import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets  # noqa
from pyzo import translate

tool_name = translate("pyzoEditorList", "Editor list")
tool_summary = "Display and manage editor tabs via a list."

Qt = QtCore.Qt


class MyQListWidget(QtWidgets.QListWidget):
    listEntryDragStart = QtCore.Signal()
    listEntryDragEnd = QtCore.Signal()
    middleButtonClicked = QtCore.Signal(QtCore.QPoint)
    doubleClicked = QtCore.Signal(QtCore.QPoint)
    keyPressed = QtCore.Signal(str)

    def dragEnterEvent(self, event):
        super().dragEnterEvent(event)
        self.listEntryDragStart.emit()

    def dropEvent(self, event):
        super().dropEvent(event)
        self.listEntryDragEnd.emit()

    def mouseDoubleClickEvent(self, event):
        super().mouseDoubleClickEvent(event)
        self.doubleClicked.emit(event.position().toPoint())

    def mousePressEvent(self, event):
        if event.button() == Qt.MouseButton.MiddleButton:
            self.middleButtonClicked.emit(event.position().toPoint())
            return
        super().mousePressEvent(event)

    def keyPressEvent(self, event):
        key = event.key()
        eventName = None
        if key in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            eventName = "enter"
        elif key == Qt.Key.Key_Delete:
            eventName = "delete"
        elif key == Qt.Key.Key_Insert:
            eventName = "insert"

        m = event.modifiers()
        if (
            m & Qt.KeyboardModifier.ControlModifier
            and m & Qt.KeyboardModifier.ShiftModifier
        ):
            if key == Qt.Key.Key_Up:
                eventName = "shift_up"
            elif key == Qt.Key.Key_Down:
                eventName = "shift_down"

        if eventName is None:
            super().keyPressEvent(event)
        else:
            self.keyPressed.emit(eventName)


class PyzoEditorList(QtWidgets.QWidget):
    """
    Mouse Control:
     - right click on a list entry opens the context menu
     - double click on a list entry activates the corresponding editor tab
     - double click below the last list entry opens a new editor tab
     - middle click on a list entry closes the corresponding editor tab
     - drag and drop to sort list entries, synchronized with the editor tabs

    Keyboard Control:
     - ENTER or RETURN activates the selected editor tab
     - DELETE will close the selected editor tab
     - INSERT opens a new editor tab
     - CTRL+SHIFT + UP/DOWN shifts the entry up/down (and the editor tab left/right)
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        # Widgets
        self._list = MyQListWidget(self)

        # Layout
        layout = QtWidgets.QVBoxLayout(self)
        self.setLayout(layout)
        layout.addWidget(self._list)

        # set margins
        margin = pyzo.config.view.widgetMargin
        layout.setContentsMargins(margin, margin, margin, margin)

        # Drag/drop
        self._indexMoveStart = None
        self._list.setDragEnabled(True)
        self._list.setDragDropMode(self._list.DragDropMode.InternalMove)
        self._list.listEntryDragStart.connect(self._onListEntryMoveStart)
        self._list.listEntryDragEnd.connect(self._onListEntryMoveEnd)

        # Context menu
        self._menu = pyzo.core.menu.EditorTabContextMenu(self, "EditorTabMenu")
        self._list.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self._list.customContextMenuRequested.connect(
            self._onCustomContextMenuRequested
        )

        self._list.middleButtonClicked.connect(self._onMiddleButtonClicked)
        self._list.doubleClicked.connect(self._onDoubleClicked)
        self._list.keyPressed.connect(self._onKeyPressed)

        pyzo.editors._tabs.tabBar().currentChanged.connect(self._onCurrentTabChanged)
        pyzo.editors._tabs.fileTabsChanged.connect(self._onFileTabsChanged)

        self.updateList()

    def updateList(self):
        tabs = pyzo.editors._tabs
        tabBar = tabs.tabBar()
        # the list widget is not cleared to keep its current scroll bar position
        i = -1  # have i initialized in case the following for loop has no iterations
        for i, tabItem in enumerate(tabs.items()):
            shortName = tabBar.tabText(i)
            longName = tabItem.filename
            shortLongInfo = []
            if tabItem.dirty:
                shortLongInfo.append(("MOD.", "MODIFIED"))
            if tabItem.pinned:
                shortLongInfo.append(("PIN.", "PINNED"))
            if tabItem.id == tabs._mainFile:
                shortLongInfo.append(("MAINF.", "MAINFILE"))
            if len(shortLongInfo) > 0:
                shortName += " [{}]".format(", ".join(sl[0] for sl in shortLongInfo))
                longName += " [{}]".format(", ".join(sl[1] for sl in shortLongInfo))
            listItem = self._list.item(i)
            if listItem is None:
                listItem = QtWidgets.QListWidgetItem()
                self._list.addItem(listItem)
            listItem.setText(shortName)
            listItem.setToolTip(longName)
        for _ in range(i + 1, self._list.count()):
            self._list.takeItem(i + 1)  # delete remaining entries
        self.selectListRow(tabs.currentIndex())

    def selectListRow(self, index):
        if index >= 0:
            self._list.setCurrentRow(index)
            self._list.scrollToItem(self._list.currentItem())

    ## Keep track of tab bar changes

    def _onCurrentTabChanged(self, index):
        self.selectListRow(index)

    def _onFileTabsChanged(self):
        self.updateList()

    ## User actions

    def _onListEntryMoveStart(self):
        item = self._list.currentItem()
        self._indexMoveStart = self._list.indexFromItem(item).row()

    def _onListEntryMoveEnd(self):
        item = self._list.currentItem()
        newIndex = self._list.indexFromItem(item).row()
        oldIndex = self._indexMoveStart
        self._indexMoveStart = None
        pyzo.editors._tabs.tabBar().moveTab(oldIndex, newIndex)
        pyzo.editors._tabs.setCurrentIndex(newIndex)

    def _onCustomContextMenuRequested(self, pos):
        """Called when context menu is clicked"""
        item = self._list.itemAt(pos)
        if item is not None:
            index = self._list.indexFromItem(item).row()
            self._menu.setIndex(index)
            self._menu.popup(self._list.viewport().mapToGlobal(pos))

    def _onDoubleClicked(self, pos):
        item = self._list.itemAt(pos)
        if item is None:
            # there was no list entry under the cursor
            pyzo.editors.newFile()
        else:
            # list entry was double clicked
            index = self._list.indexFromItem(item).row()
            pyzo.editors._tabs.setCurrentIndex(index)

    def _onMiddleButtonClicked(self, pos):
        item = self._list.itemAt(pos)
        if item is not None:
            index = self._list.indexFromItem(item).row()
            pyzo.editors._tabs.tabCloseRequested.emit(index)

    def _onKeyPressed(self, eventName):
        index = self._list.currentRow()
        if index >= 0:
            if eventName == "enter":
                pyzo.editors._tabs.setCurrentIndex(index)
            elif eventName == "delete":
                pyzo.editors._tabs.tabCloseRequested.emit(index)
            elif eventName.startswith("shift_"):
                oldIndex = index
                newIndex = index
                if eventName == "shift_up":
                    newIndex = index - 1
                elif eventName == "shift_down":
                    newIndex = index + 1
                if newIndex != oldIndex and 0 <= newIndex < self._list.count():
                    pyzo.editors._tabs.tabBar().moveTab(oldIndex, newIndex)
                    pyzo.editors._tabs.setCurrentIndex(newIndex)
        if eventName == "insert":
            pyzo.editors.newFile()



import os


# Automatically scale along on HDPI displays. See issue #531 and e.g.
# https://wiki.archlinux.org/index.php/HiDPI#Qt_5
if "QT_AUTO_SCREEN_SCALE_FACTOR" not in os.environ:
    os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"

# Fix Qt now showing a window on MacOS Big Sur
os.environ["QT_MAC_WANTS_LAYER"] = "1"



import os


def count_lines(filename):
    # Get text
    with open(filename, "rt", encoding="utf-8") as fd:
        text = fd.read()

    # Init counts: code, docstring, comments, whitespace
    count1, count2, count3, count4 = 0, 0, 0, 0

    inDocstring = False
    for line in text.splitlines():
        line = line.strip()
        if '"' * 3 in line:
            inDocstring = not inDocstring
        if not line:
            count4 += 1
        elif inDocstring:
            count2 += 1
        elif line.startswith(("#", "%")):
            count3 += 1
        else:
            count1 += 1

    # Done
    return count1, count2, count3, count4


if __name__ == "__main__":
    # Get path of yoton
    path = os.path.dirname(os.path.abspath(__file__))
    path = os.path.split(path)[0]

    # Init files
    files = []
    # Get files in root
    for fname in os.listdir(path):
        files.append(fname)
    # Get files in channels
    for fname in os.listdir(os.path.join(path, "channels")):
        fname = os.path.join("channels", fname)
        files.append(fname)

    N1, N2, N3, N4 = 0, 0, 0, 0
    for fname in files:
        if not fname.endswith(".py"):
            continue
        n1, n2, n3, n4 = count_lines(os.path.join(path, fname))
        N1 += n1
        N2 += n2
        N3 += n3
        N4 += n4
        print("%i lines in %s" % (n1, fname))

    print("yoton has %i lines in its source" % (N1 + N2 + N3 + N4))
    print("yoton has %i lines of code" % N1)
    print("yoton has %i lines of docstring" % N2)
    print("yoton has %i comment lines" % N3)
    print("yoton has %i lines of whitespace" % N4)



import os
import re
import shutil
import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets  # noqa
from pyzo import translate

tool_name = translate("snippets", "Snippets")
tool_summary = "source code and text snippets"

Qt = QtCore.Qt


class PyzoSnippets(QtWidgets.QWidget):
    _searchOptions = {  # value: config_name, default_value, label
        "name": ("searchName", True, translate("snippets", "Search name")),
        "description": (
            "searchDescription",
            True,
            translate("snippets", "Search description"),
        ),
        "code": ("searchCode", True, translate("snippets", "Search code")),
    }

    def __init__(self, parent=None):
        super().__init__(parent)

        toolId = self.__class__.__name__.lower()
        self._config = pyzo.config.tools[toolId]

        self._loadSnippets()

        # setup layout
        self._layout = QtWidgets.QVBoxLayout()
        self.setLayout(self._layout)
        self._hlayout = QtWidgets.QHBoxLayout()
        self._layout.addLayout(self._hlayout)

        # create tree widget
        self._tree = QtWidgets.QTreeWidget()
        self._tree.setHeaderHidden(True)
        self._tree.itemSelectionChanged.connect(self._onSelectionChanged)
        self._tree.itemDoubleClicked.connect(self._onItemDoubleClicked)
        self._layout.addWidget(self._tree)

        # create context menu for tree widget
        self._menu = pyzo.core.menu.Menu(self, "Snippets")
        self._menu.addItem(
            translate(
                "snippets",
                "Insert into editor ::: Insert code snippet at the cursor in the current editor",
            ),
            pyzo.icons.paste_plain,
            self._onMenuChosen,
            "insert",
        )
        self._menu.addItem(
            translate("snippets", "Edit ::: Open snippet file in Pyzo"),
            pyzo.icons.application_edit,
            self._onMenuChosen,
            "edit",
        )
        self._tree.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self._tree.customContextMenuRequested.connect(
            self._onCustomContextMenuRequested
        )

        # create search line edit
        self._searchText = QtWidgets.QLineEdit()
        self._searchText.textChanged.connect(self._updateTreeWidget)
        self._hlayout.addWidget(self._searchText)

        # create menu button
        btn = QtWidgets.QToolButton()
        self._menuButton = btn
        self._hlayout.addWidget(btn)
        btn.setIcon(pyzo.icons.application_view_list)
        btn.setIconSize(QtCore.QSize(16, 16))
        btn.setPopupMode(btn.ToolButtonPopupMode.InstantPopup)
        btn.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        btn._menu = QtWidgets.QMenu()
        btn.setMenu(btn._menu)
        btn.setToolTip(
            'Type "SNIP." in the editor to insert snippets via autocompletion\n'
            "or double-click an entry to insert it at the cursor position.\n"
            'Right-click an entry and select "Edit" to open the file of the snippet.\n'
            "Open the example snippet file for more information.\n"
        )
        btn.pressed.connect(self._onSearchOptionsPress)
        btn._menu.triggered.connect(self._onSearchOptionMenuTiggered)
        self._onSearchOptionsPress()  # create menu now and init default config values

        # set margins
        margin = pyzo.config.view.widgetMargin
        self._layout.setContentsMargins(margin, margin, margin, margin)

        self._updateTreeWidget()
        self._updateSearchPlaceHolderText()

    def _onSearchOptionsPress(self):
        """create the menu for the button

        Do each time to make sure the check boxes have correct values.
        """
        menu = self._menuButton._menu
        menu.clear()

        for name, default, label in self._searchOptions.values():
            action = menu.addAction(label)
            action._what = name
            action.setCheckable(True)
            action.setChecked(self._config.setdefault(name, default))

        menu.addSeparator()

        menuEntries = [
            ("reloadSnippets", translate("snippets", "Reload snippets")),
            (
                "openSnippetsDir",
                translate("snippets", "Open snippets directory outside Pyzo"),
            ),
        ]

        for name, label in menuEntries:
            action = menu.addAction(label)
            action._what = name

    def _onSearchOptionMenuTiggered(self, action):
        name = action._what
        if name in ("searchName", "searchDescription", "searchCode"):
            self._config[name] ^= True
            self._updateTreeWidget()
            self._updateSearchPlaceHolderText()
        elif name == "reloadSnippets":
            self._loadSnippets()
            self._updateTreeWidget()
        elif name == "openSnippetsDir":
            pyzo.util.open_directory_outside_pyzo(self._snippetsDir)

    def _updateSearchPlaceHolderText(self):
        searchKeys = []
        for name, (configName, default, label) in self._searchOptions.items():
            if getattr(self._config, configName):
                searchKeys.append(name)
        if searchKeys:
            t = "search: " + ", ".join(searchKeys)
        else:
            t = "[Please activate at least one search option]"
        self._searchText.setPlaceholderText(t)

    def _onCustomContextMenuRequested(self, pos):
        """Called when context menu is clicked"""

        # get item that was clicked on
        item = self._tree.itemAt(pos)
        if item is None or not hasattr(item, "_snip"):
            return

        self._menu.popup(self._tree.viewport().mapToGlobal(pos + QtCore.QPoint(3, 3)))

    def _onMenuChosen(self, arg):
        item = self._tree.selectedItems()[0]
        snip = getattr(item, "_snip", None)
        if snip is None:
            return

        if arg == "insert":
            ed = pyzo.editors.getCurrentEditor()
            if ed:
                cur = ed.textCursor()
                self._insertSnippet(cur, snip)
                ed.setFocus()
        elif arg == "edit":
            res = pyzo.editors.loadFile(item._snip["filepath"])
            linenr = item._snip["indLine"] + 1
            if res:
                ed = res._editor
                ed.gotoLine(linenr)
                ed.setFocus()

    def _addTreeElems(self, parentItem, snippetsPath):
        snip = self._snippetsFlatTreeFiltered[snippetsPath]
        if isinstance(snip, list):
            for name in snip:
                item = QtWidgets.QTreeWidgetItem(parentItem, [name])
                item.setExpanded(True)
                snippetsPath2 = snippetsPath + (name,)
                self._addTreeElems(item, snippetsPath2)
                snip2 = self._snippetsFlatTreeFiltered[snippetsPath2]
                if isinstance(snip2, dict):
                    item._snip = snip2
                    tt = "filepath: {}\nline number: {}".format(
                        snip2["filepath"], snip2["indLine"] + 1
                    )
                    if snip2["description"]:
                        tt += "\n\n{}".format(snip2["description"])
                    item.setToolTip(0, tt)

    def _onSelectionChanged(self):
        selectedItems = self._tree.selectedItems()
        if selectedItems:
            (item,) = selectedItems
            snip = getattr(item, "_snip", None)
        else:
            # nothing selected or empty list
            snip = None
        self.showHelpForSnip(snip)

    def _onItemDoubleClicked(self, item, column):
        snip = getattr(item, "_snip", None)
        if snip:
            ed = pyzo.editors.getCurrentEditor()
            if ed:
                cur = ed.textCursor()
                self._insertSnippet(cur, snip)
                ed.setFocus()

    def _filterTree(self):
        searchKeys = []
        for name, (configName, default, label) in self._searchOptions.items():
            if getattr(self._config, configName):
                searchKeys.append(name)
        needle = self._searchText.text().casefold()

        flatTree = {(): []}
        for treePath, snip in self._snippetsFlatTree.items():
            if isinstance(snip, dict):
                if any(needle in snip[k].casefold() for k in searchKeys):
                    flatTree[treePath] = snip
                    while True:
                        name = treePath[-1]
                        treePath = treePath[:-1]
                        if treePath in flatTree:
                            flatTree[treePath].append(name)
                            break
                        else:
                            flatTree[treePath] = [name]
        self._snippetsFlatTreeFiltered = flatTree

    def _updateTreeWidget(self):
        self._tree.clear()
        self._filterTree()
        self._addTreeElems(self._tree, ())

    def getAutocompleteNames(self, name):
        # name has to be "SNIP" or "SNIP.something" or "SNIP.something.otherthing" ...
        # returns None or a list of names
        namesOrSnippet = self._snippetsNameLookUp.get(name, [])
        if isinstance(namesOrSnippet, list) and namesOrSnippet:
            return namesOrSnippet
        return None

    def isSnipName(self, name):
        return name.startswith("SNIP.")

    def showHelpForSnip(self, nameOrSnip):
        helpTool = pyzo.toolManager.getTool("pyzointeractivehelp")
        if helpTool:
            if isinstance(nameOrSnip, dict):
                snip = nameOrSnip
            else:
                snip = self._snippetsNameLookUp.get(nameOrSnip, None)
            if isinstance(snip, dict):
                helpTool.helpForCodeSnippet(snip)
            else:
                helpTool.helpForCodeSnippet(None)

    def _insertSnippet(self, textCursor, snip):
        cur = textCursor
        if cur.hasSelection():
            posInBlock = cur.selectionStart() - cur.block().position()
        else:
            posInBlock = cur.positionInBlock()

        textBefore = cur.block().text()[:posInBlock]
        code = snip["code"]
        if textBefore.isspace():
            lines = code.split("\n")
            code = "\n".join(lines[:1] + [(textBefore + s).rstrip() for s in lines[1:]])
        cur.insertText(code)

    def applyAutoCompletion(self, textCursor):
        cur = textCursor
        line = cur.block().text()
        posInBlock = cur.positionInBlock()

        mo = re.fullmatch(r"(.*?\s+|^)(SNIP(?:\.\w+)+)", line[:posInBlock])
        if mo:
            name = mo[2]
            snip = self._snippetsNameLookUp.get(name, None)
            if isinstance(snip, dict):
                cur.setPosition(cur.position() - len(name), cur.MoveMode.KeepAnchor)
                self._insertSnippet(cur, snip)

    def _loadSnippets(self):
        if not hasattr(self._config, "customSnippetsPath"):
            self._config.customSnippetsPath = ""

        if self._config.customSnippetsPath:
            self._snippetsDir = self._config.customSnippetsPath
        else:
            self._snippetsDir = os.path.join(pyzo.appDataDir, "snippets")

        if not self._config.customSnippetsPath:
            if os.path.isdir(os.path.dirname(self._snippetsDir)) and not os.path.isdir(
                self._snippetsDir
            ):
                os.makedirs(self._snippetsDir)
                fpSrc = os.path.join(
                    pyzo.pyzoDir, "resources", "code_examples", "snippets_example.py"
                )
                fpDst = os.path.join(self._snippetsDir, "snippets_example.py")
                shutil.copyfile(fpSrc, fpDst)

        flatTree = {(): []}
        if os.path.isdir(self._snippetsDir):
            stack = [()]
            while stack:
                treePath = stack.pop()
                dirpath = os.path.join(self._snippetsDir, *treePath)
                for s in os.listdir(dirpath):
                    fp = os.path.join(dirpath, s)
                    dirNames = []
                    if os.path.isdir(fp):
                        if _NAME_PATTERN.fullmatch(s):
                            dirNames.append(s)
                        else:
                            print(
                                f'WARNING: ignoring folder "{fp}" because its name is not valid'
                            )
                    elif os.path.isfile(fp):
                        if s.endswith(".py"):
                            for snip in _extractSnippetsFromFile(fp):
                                name = snip["name"]
                                treePath2 = treePath + (name,)

                                if treePath2 in flatTree:
                                    print(
                                        f'WARNING: ignoring snippet "{name}" in file "{fp}" because the name is already used in another snippet'
                                    )
                                    continue

                                flatTree[treePath2] = snip
                                while True:
                                    name2 = treePath2[-1]
                                    treePath2 = treePath2[:-1]
                                    if treePath2 in flatTree:
                                        flatTree[treePath2].append(name2)
                                        break
                                    else:
                                        flatTree[treePath2] = [name2]

                        else:
                            print(
                                f'WARNING: ignoring file "{fp}" because name has no ending ".py"'
                            )

                    for s in dirNames:
                        treePath2 = treePath + (s,)
                        if treePath2 in flatTree:
                            fp = os.path.join(dirpath, s)
                            print(
                                f'WARNING: ignoring folder "{fp}" because there is already a snippet with the same name'
                            )
                        else:
                            stack.append(treePath2)

        self._snippetsFlatTree = flatTree
        self._snippetsNameLookUp = {
            ".".join(("SNIP",) + treePath): v for treePath, v in flatTree.items()
        }


_NAME_PATTERN = re.compile(r"[a-zA-Z_][a-zA-Z_0-9]*")
_SNIP_TITLE_PATTERN = re.compile(r"^\s*\#\#\s*SNIP:\s*([a-zA-Z_][a-zA-Z_0-9]*)\s*$")


def _stripLines(lines):
    """removes empty or whitespace-only lines at the start and end of list lines"""
    i = 0
    for i, line in enumerate(lines):
        if line.strip():
            break
    lines = lines[i:]
    i = 0
    for i, line in enumerate(lines[::-1]):
        if line.strip():
            break
    lines = lines[: len(lines) - i]
    return lines


def _extractSnippetsFromFile(filepath):
    snippets = []

    try:
        with open(filepath, "rt", encoding="utf-8") as fd:
            lines = fd.read().split("\n")
    except Exception:
        print("WARNING: could not read snippets from file:", repr(filepath))
    else:
        curName = None
        curStartLineInd = None
        curDescriptionLines = []
        curCodeLines = []
        for indLine, line in enumerate(lines + ["## SNIP: END_OF_FILE"]):
            mo = _SNIP_TITLE_PATTERN.fullmatch(line)
            if mo:
                if curName is not None:
                    curCodeLines = _stripLines(curCodeLines)
                    if len(curCodeLines) > 1:
                        # add a linebreak if there is more than 1 line
                        curCodeLines.append("")
                    snippets.append(
                        {
                            "name": curName,
                            "indLine": curStartLineInd,
                            "description": "\n".join(curDescriptionLines),
                            "code": "\n".join(curCodeLines),
                            "filepath": filepath,
                        }
                    )
                curName = mo[1]
                curStartLineInd = indLine
                curDescriptionLines = []
                curCodeLines = []
            else:
                if not curCodeLines:
                    line2 = line.lstrip()
                    if line2.startswith("#"):
                        line2 = line2[1:]
                        if line2[:1] == " ":
                            line2 = line2[1:]
                        curDescriptionLines.append(line2)
                    else:
                        curCodeLines.append(line)
                else:
                    curCodeLines.append(line)

    return snippets



import sys
import time
import struct
import socket

# import select  # to determine wheter a socket can receive data
if sys.platform.startswith("java"):  # Jython
    from select import cpython_compatible_select as select, error as SelectErr
else:
    from select import select, error as SelectErr

from yoton.misc import bytes


## Constants

# Error code for interruptions
try:
    from errno import EINTR
except ImportError:
    EINTR = None

# Queue size
BUF_MAX_LEN = 10000

# Define buffer size. For recv 4096 or 8192 chunk size is recommended.
# For sending, in principle as large as possible, but prevent too much
# message copying.
BUFFER_SIZE_IN = 2**13
BUFFER_SIZE_OUT = 1 * 2**20

# Reserved slots (slot 0 does not exist, so we can test "if slot: ")
# The actual slots start counting from 8.
# SLOT_TEST = 3 -> Deprecated
SLOT_CONTEXT = 2
SLOT_DUMMY = 1

# Struct header packing
HEADER_FORMAT = "<7Q"
HEADER_SIZE = struct.calcsize(HEADER_FORMAT)

# Constant for control bytes
(CONTROL_BYTES,) = struct.unpack("<Q", b"YOTON".ljust(8))


## Helper functions


def can_send(s, timeout=0.0):
    """can_send(bsd_socket, timeout=0.0)

    Given a socket, uses select() to determine whether it can be used
    for sending. This function can deal with system interrupts.

    """
    while True:
        try:
            can_recv, can_send, tmp = select([], [s], [], timeout)
            break
        except SelectErr:
            # Skip errors caused by interruptions.
            type, value, tb = sys.exc_info()
            del tb
            if value.args[0] != EINTR:
                raise
    return bool(can_send)


def can_recv(s, timeout=0.0):
    """can_recv(bsd_socket, timeout=0.0)

    Given a socket, uses select() to determine whether it can be used
    for receiving. This function can deal with system interrupts.

    """
    while True:
        try:
            can_recv, can_send, tmp = select([s], [], [], timeout)
            break
        except SelectErr:
            # Skip errors caused by interruptions.
            type, value, tb = sys.exc_info()
            del tb
            if value.args[0] != EINTR:
                raise
    return bool(can_recv)


def send_all(s, text, shutdown_after_sending=True):
    """send_all(socket, text, shutdown_after_sending=True)

    Send all text to the socket. Used during handshaking and in
    the clientserver module.

    If shutdown_after_sending, the socket is shut down. Some protocols
    rely on this.

    It is made sure that the text ends with a CRLF double-newline code.

    """

    # Ensure closing chars
    if not text.endswith("\r\n"):
        text += "\r\n"

    # Make bytes
    bb = text.encode("utf-8")

    # Send all bytes
    try:
        s.sendall(bb)  # -> n
    except socket.error:
        return -1  # Socket closed down badly

    # Shutdown connection nicely from here
    if shutdown_after_sending:
        try:
            s.shutdown(socket.SHUT_WR)
        except socket.error:
            pass


def recv_all(s, timeout=-1, end_at_crlf=True):
    """recv_all(socket, timeout=-1, end_at_crlf=True)

    Receive text from the socket (until socket receiving is shut down).
    Used during handshaking and in the clientserver module.

    If end_at_crlf, a message is also ended at a CRLF double-newline code,
    and a shutdown is not necessary. This takes a tiny bit longer.

    """

    # Init parts (start with one byte, such that len(parts) is always >= 2
    parts = [b" "]

    # Determine number of bytes to get per recv
    nbytesToGet = BUFFER_SIZE_IN
    if end_at_crlf:
        nbytesToGet = 1

    # Set end bytes
    end_bytes = b"\r\n"

    # Set max time
    if timeout <= 0:
        timeout = 2**32
    maxtime = time.time() + timeout

    # Receive data
    while True:
        # Receive if we can
        if can_recv(s):
            # Get part
            try:
                part = s.recv(nbytesToGet)
                parts.append(part)
            except socket.error:
                return None  # Socket closed down badly

            # Detect end by shutdown (EOF)
            if not part:
                break

            # Detect end by \r\n
            if end_at_crlf and (parts[-2] + parts[-1]).endswith(end_bytes):
                break

        else:
            # Sleep
            time.sleep(0.01)

            # Check time
            if time.time() > maxtime:
                bb = b"".join(parts[1:])
                return bb.decode("utf-8", "ignore")

    # Combine parts (discarded first (dummy) part)
    bb = b"".join(parts[1:])

    # Try returning as Unicode
    try:
        return bb.decode("utf-8", "ignore")
    except UnicodeError:
        return "<UnicodeError>"


## Package class


class Package(object):
    """Package(data, slot, source_id, source_seq, dest_id, dest_seq, recv_seq)

    Represents a package of bytes to be sent from one Context instance
    to another. A package consists of a header and the encoded message.

    To make this class as fast as reasonably possible, its interface
    is rather minimalistic and few convenience stuff is implemented.

    Parameters
    ----------
    data : bytes
        The message itself.
    slot : long
        The slot at which the package is directed. The integer is a hash of
        a string slot name.
    source_id : long
        The id of the context that sent this package.
    source_seq : long
        The sequence number of this package, counted at the sending context.
        Together with source_id, this fully identifies a package.
    dest_id : long (default 0)
        The id of the package that this package replies to.
    dest_seq : long (default 0)
        The sequence number of the package that this package replies to.
    recv_seq : long (default 0)
        The sequence number of this package counted at the receiving context.
        This is used to synchronize channels.

    When send, the header is composed of: four control bytes, the slot,
    the source_id, source_seq, dest_id and dest_seq.

    Notes
    -----
    A package should always have content. Packages without content are only
    used for low-level communication between two ContextConnection instances.
    The source_seq is then used as the signal. All other package attributes
    are ignored.

    """

    # The __slots__ makes instances of this class consume < 20% of memory
    # Note that this only works for new style classes.
    # This is important because many packages can exist at the same time
    # if a receiver can't keep up with a sender. Further, although Python's
    # garbage collector collects the objects after they're "consumed",
    # it does not release the memory, because it hopes to reuse it in
    # an efficient way later.
    __slots__ = [
        "_data",
        "_dest_id",
        "_dest_seq",
        "_recv_seq",
        "_slot",
        "_source_id",
        "_source_seq",
    ]

    def __init__(self, data, slot, source_id, source_seq, dest_id, dest_seq, recv_seq):
        self._data = data
        self._slot = slot
        #
        self._source_id = source_id
        self._source_seq = source_seq
        self._dest_id = dest_id
        self._dest_seq = dest_seq
        self._recv_seq = recv_seq

    def parts(self):
        """parts()

        Get list of bytes that represents this package.
        By not concatenating the header and content parts,
        we prevent unnecessary copying of data.

        """

        # Obtain header
        header = struct.pack(
            HEADER_FORMAT,
            CONTROL_BYTES,
            self._slot,
            self._source_id,
            self._source_seq,
            self._dest_id,
            self._dest_seq,
            len(self._data),
        )

        # Return header and message
        return header, self._data

    def __str__(self):
        """Representation of the package. Mainly for debugging."""
        return "At slot %i: %s" % (self._slot, repr(self._data))

    @classmethod
    def from_header(cls, header):
        """from_header(header)

        Create a package (without data) from the header of a message.
        Returns (package, data_length). If the header is invalid (checked
        using the four control bytes) this method returns (None, None).

        """
        # Unpack
        tmp = struct.unpack(HEADER_FORMAT, header)
        CTRL, slot, source_id, source_seq, dest_id, dest_seq, L = tmp
        # Create package
        p = Package(None, slot, source_id, source_seq, dest_id, dest_seq, 0)
        # Return
        if CTRL == CONTROL_BYTES:
            return p, L
        else:
            return None, None


# Constant Package instances (source_seq represents the signal)
PACKAGE_HEARTBEAT = Package(bytes(), SLOT_DUMMY, 0, 0, 0, 0, 0)
PACKAGE_CLOSE = Package(bytes(), SLOT_DUMMY, 0, 1, 0, 0, 0)



import yoton
import threading
import time


class Sender(threading.Thread):
    def run(self):
        for i in range(100):
            time.sleep(0.02)
            q.push("haha! " + repr(self))
        q.push("stop")


class Receiver(threading.Thread):
    def run(self):
        M = []
        L = []
        while True:
            L.append(len(q))
            m = q.pop(True)
            # time.sleep(0.02)
            if m == "stop":
                break
            else:
                M.append(m)

        self.M = M
        self.avSize = float(sum(L)) / len(L)

    def show(self):
        M = self.M
        nrs = {}
        for m in M:
            nr = m.split("-", 1)[1].split(",", 1)[0]
            if nr not in nrs:
                nrs[nr] = 0
            nrs[nr] += 1

        print("received %i messages" % len(M))
        print("average queue size was %1.2f" % self.avSize)
        for nr in nrs:
            print("  from %s received %i" % (nr, nrs[nr]))


q = yoton.misc.TinyPackageQueue(10, 1000)

S = []
for i in range(10):
    t = Sender()
    t.start()
    S.append(t)

R = []
for i in range(3):
    t = Receiver()
    t.start()
    R.append(t)

for r in R:
    r.join()
    r.show()
